/**
	@file token.d
	@brief スクリプトを解析する。
	@author senzogawa
**/
module nscr.token;

private import std.ctype;
debug(uttoken) private import std.windows.charset;
private import std.file;
private import std.string;
private import std.conv;
private import core.stdc.stdio; // printf

/// トークンの種別
enum TokenType
{
	STRING,            ///< 文字列。
	SPACE,             ///< 空白。タブも含める。
	COMMENT,           ///< コメント
	LITERAL,           ///< 文字列リテラル
	SPECIAL_LABEL,     ///< 特殊ラベル。
	LINETOP_LABEL,     ///< 行頭ラベル。
	GENERIC_LABEL,     ///< 一般ラベル。引数として使うやつ。
	NUMBER_VARIABLE,   ///< 数字変数
	STRING_VARIABLE,   ///< 文字変数
	ARRAY_VARIABLE,    ///< 配列変数
	DEFINE_COMMAND,    ///< 定義節命令
	EXECUTE_COMMAND,   ///< 実行節命令
	COMMON_COMMAND,    ///< 共通命令
	COLOR,             ///< 色値
	NUMBER,            ///< 数値
	SYMBOL,            ///< シンボル
	SPECIAL_CHARACTER, ///< 特殊文字
	OPERATOR,          ///< 演算子。modは含まない。
	RUBY,              ///< ルビ
	TAG,               ///< タグ
	TEXT_BUTTON,       ///< テキストボタン
	PARENTHESIS_BEGIN, ///< 小括弧開始
	PARENTHESIS_END,   ///< 小括弧終了
	BRACKET_BEGIN,     ///< 大括弧開始
	BRACKET_END,       ///< 大括弧終了
	ASCII              ///< ASCIIコード
}

bool isAcceptableAsParam(TokenType type){
    switch(type){
    case TokenType.STRING:
    case TokenType.COMMENT:
    case TokenType.SPECIAL_CHARACTER:
    case TokenType.RUBY:
    case TokenType.TAG:
    case TokenType.TEXT_BUTTON:
    case TokenType.BRACKET_BEGIN:
    case TokenType.BRACKET_END:
    case TokenType.ASCII:
        return false;
    default:
        return true;
    }
}

/// トークンの情報
struct Token
{
	void opCall(string value, TokenType type){
		this.value = value;
		this.type = type;
	}
	void opCall(char[] value, TokenType type){
	    this.value = value.idup;
		this.type = type;
	}
	void opCall(char value, TokenType type){
		this.value ~= value;
		this.type = type;
	}
	string value = "";
	TokenType type = TokenType.STRING;
	static Token init = {value.init, type.init};
	void dump(){
		printf("%d:(%d):%.*s", type, value.length, value);
	}
}


TokenType[string] RESERVED_COMMANDS;


/// 予約命令をファイルから読み込む。
void loadReservedCommands(string filename){
	string[] lines = splitlines(cast(string)read(filename));
	string[] factors;
	foreach(string line; lines){
		factors = split(line, ",");
		if(factors.length == 1){
			RESERVED_COMMANDS[factors[0]] = TokenType.COMMON_COMMAND;
		}
		else if(factors.length == 2){
			RESERVED_COMMANDS[factors[0]] = getCommandType(factors[1]);
		}
		else{
			throw new Exception("bad format the description of reserved commands");
		}
	}
}


/// 命令の種別を数値化する。
TokenType getCommandType(string command){
	TokenType type;
	switch(command){
	case "F":
		type = TokenType.DEFINE_COMMAND;
		break;
	case "FE":
		type = TokenType.COMMON_COMMAND;
		break;
	case "E":
		type = TokenType.EXECUTE_COMMAND;
		break;
	default:
		type = TokenType.COMMON_COMMAND;
		break;
	}
	return type;
}


/// 各行のトークンを切り出す。
Token[][] parseToTokenLines(string[] lines){
	Token[][] tokenlines;
	tokenlines.length = lines.length;
	int lineno;
	try{
		for(lineno=0; lineno<lines.length; lineno++){
			tokenlines[lineno] = parseToTokens(lines[lineno]);
		}
	}
	catch(Exception e){
		throw new Exception(to!string(lineno) ~ ": " ~ e.toString());
	}
	return tokenlines;
}


/// 1行中のトークン群を切り出す。
Token[] parseToTokens(string line){
	Token[] tokens;
	Token token;
	Token[1] token_buf;
	char[1] ch;
	// 終端判定
	int iseos(int position){
		return (position == line.length-1);
	}
	for(int pos=0; pos<line.length; pos++){
		ch[0] = line[pos];
		switch(ch){
		// SPACE
		case " ":
		case "\t":
			token(parseSpace(line, pos), TokenType.SPACE);
			break;
		// COMMENT
		case ";":
			token(line[pos..$], TokenType.COMMENT);
			pos = line.length; // 終了させる。
			break;
		// LITERAL
		case "\"":
			token(parseLiteral(line, pos), TokenType.LITERAL);
			break;
		// LABEL or OPERATOR
		case "*":
			if(iseos(pos)){
				if(tokens.length > 0){
					// 先頭が空白で、かつ空白以外が無い場合のみ行頭。
					if(!(tokens[0].type == TokenType.SPACE && tokens.length == 1)){
						// 行頭のラベルではない。
						token(line[pos], TokenType.GENERIC_LABEL);
						break;
					}
				}
				// 末尾の「*」は無名ラベルとみなす。
				token(line[pos], TokenType.LINETOP_LABEL);
				break;
			}
			if(isSymbolCap(line[pos+1])){
				// ラベルとして処理する。
				pos++;
				string symbol = parseSymbol(line, pos);
				if(tokens.length > 0){
					// 先頭が空白で、かつ空白以外が無い場合のみ行頭。
					if(!(tokens[0].type == TokenType.SPACE && tokens.length == 1)){
						// 行頭のラベルではない。
						token(ch ~ symbol, TokenType.GENERIC_LABEL);
						break;
					}
				}
				// 行頭ラベルとして処理。
				switch(symbol){
				case "define":
				case "start":
				case "customsel":
					token(ch ~ symbol, TokenType.SPECIAL_LABEL);
					break;
				default:
					token(ch ~ symbol, TokenType.LINETOP_LABEL);
					break;
				}
			}
			else{
				// 演算子
				token(line[pos], TokenType.OPERATOR);
			}
			break;
		// VARIABLE
		case "%":
			if(iseos(pos)){
				// 末尾に「%」はありえないので、フォーマット不正
				throw new Exception("input symbol or number to line bottom");
			}
			if(iseos(pos+1) && line[pos+1] == '%'){
				// 末尾に「%%」はありえないので、フォーマット不正
				throw new Exception("input symbol or number to line bottom");
			}
			token(parseIntegerVariable(line, pos), TokenType.NUMBER_VARIABLE);
			break;
		case "$":
			if(iseos(pos)){
				// 末尾に「$」はありえないので、フォーマット不正
				throw new Exception("input symbol or number to line bottom");
			}
			if(iseos(pos+1) && line[pos+1] == '%'){
				// 末尾に「$%」はありえないので、フォーマット不正
				throw new Exception("input symbol or number to line bottom");
			}
			token(parseStringVariable(line, pos), TokenType.STRING_VARIABLE);
			break;
		case "?":
			if(iseos(pos)){
				// 末尾に「?」はありえないので、フォーマット不正
				throw new Exception("input symbol or number to line bottom");
			}
			if(iseos(pos+1) && line[pos+1] == '%'){
				// 末尾に「?%」はありえないので、フォーマット不正
				throw new Exception("input symbol or number to line bottom");
			}
			token(parseArrayVariable(line, pos), TokenType.ARRAY_VARIABLE);
			break;
		// COLOR
		case "#":
			pos++;
			token(ch ~ parseXDigit(line, pos), TokenType.COLOR);
			if(token.value.length != 7){
				throw new Exception("the format of color is #[0-9a-fA-F]{6}");
			}
			break;
		// SPECIAL_CHARACTER, OPERATOR
		case "!":
			if(iseos(pos)){
				// 末尾に「!」はありえないので、フォーマット不正
				throw new Exception("input 'd', 's', 'sd', 'w' to line bottom");
			}
			pos++;
			if(line[pos] == '='){
				// !=の場合の場合はここで確定。
				token(line[pos-1..pos+1], TokenType.OPERATOR);
				break;
			}
			token(ch ~ parseSymbol(line, pos), TokenType.SPECIAL_CHARACTER);
			// 汚いが、とりあえず。
			switch(token.value){
			case "!d":
			case "!s":
			case "!sd":
			case "!w":
				break;
			default:
				throw new Exception("input 'd', 's', 'sd', 'w' follow '!'");
			}
			break;
		case "\\":
		case "@":
			token(line[pos], TokenType.SPECIAL_CHARACTER);
			break;
		case "_":
			if(iseos(pos) || !isSymbolChar(line[pos+1])){
				token(line[pos], TokenType.SPECIAL_CHARACTER);
			}
			else{
				// 後ろにシンボル文字が続く時はシンボルになる。
				token(parseSymbol(line, pos), TokenType.SYMBOL);
			}
			break;
		// OPERATOR
		case "-":
		case "+":
		case "=":
		case "&":
		case ":":
			token(line[pos], TokenType.OPERATOR);
			break;
		// OPERATOR, SPECIAL_CHARACTER
		case "/":
			// 行末の場合は改行を打ち消す特殊文字
			if(iseos(pos)){
				token(line[pos], TokenType.SPECIAL_CHARACTER);
			}
			else{
				token(line[pos], TokenType.OPERATOR);
			}
			break;
		// RUBY , PARENTHESIS_BEGIN
		case "(":
			if(iseos(pos)){
				// 末尾に「(」はありえないので、フォーマット不正
				throw new Exception("not found ')'.");
			}
			if(isRubyString(line, pos)){
				token(parseMarkuppedString(line, pos, '(', ')'), TokenType.RUBY);
			}
			else{
				token(line[pos], TokenType.PARENTHESIS_BEGIN);
			}
			break;
		// PARENTHESIS_END
		case ")":
			token(line[pos], TokenType.PARENTHESIS_END);
			break;
		// TAG, BRACKET_BEGIN
		case "[":
			if(tokens.length == 0 || isOnlySpace(tokens)){
				token(parseMarkuppedString(line, pos, '[', ']'), TokenType.TAG);
			}
			else{
				token(line[pos], TokenType.BRACKET_BEGIN);
			}
			break;
		// BRACKET_END
		case "]":
			token(line[pos], TokenType.BRACKET_END);
			break;
		// TEXT_BUTTON, OPERATOR
		case "<":
			if(isTextButton(line, pos)){
				token(parseMarkuppedString(line, pos, '<', '>'), TokenType.TEXT_BUTTON);
			}
			else{
				token(line[pos], TokenType.OPERATOR);
			}
			break;
		// NUMBER, SYMBOL, STRING
		default:
			if(isdigit(line[pos])){
				token(parseNumber(line, pos), TokenType.NUMBER);
			}
			else if(isSymbolCap(line[pos])){
				token(parseSymbol(line, pos), TokenType.SYMBOL);
			}
			else if(isascii(line[pos])){
				token(line[pos], TokenType.ASCII);
			}
			else{
				token(parseString(line, pos), TokenType.STRING);
			}
			break;
		}
		if(token.type == TokenType.SYMBOL){
			if(token.value in RESERVED_COMMANDS){
			    token.type = RESERVED_COMMANDS[token.value];
			}
		}
		token_buf[0] = token;
		tokens ~= token_buf;
		token = token.init;
	}
	return tokens;
}


/// 空白列を解析する。
/**
	@param str 文字列。([ ]|\t).*
	@param pos    位置。終了後は終端の一つ後が格納されている。
	@return リテラル。([ ]|\t)+
**/
string parseSpace(string str, ref int pos){
	assert(str.length > pos);
	int start_pos = pos;
	for(; pos<str.length; pos++){
		if(str[pos] != ' ' && str[pos] != '\t'){
			pos--;
			break;
		}
	}
	if(pos > str.length-1){
		pos = str.length-1;
	}
	return str[start_pos..pos+1];
}


/// シンボルを解析する。
/**
	@param str 文字列。[_a-zA-Z].*
	@param pos    位置。終了後は終端の一つ後が格納されている。
	@return シンボル。[_a-zA-Z][_a-zA-Z0-9]*
**/
string parseSymbol(string str, ref int pos){
	assert(str.length > pos);
	assert(isSymbolCap(str[pos]));
	int start_pos = pos;
	pos++;
	for(; pos<str.length; pos++){
		if(!isSymbolChar(str[pos])){
			pos--;
			break;
		}
	}
	if(pos > str.length-1){
		pos = str.length-1;
	}
	return str[start_pos..pos+1];
}


/// リテラルを解析する。
/**
	@param str 文字列。"[^"]*".*
	@param pos    位置。終了後は終端の一つ後が格納されている。
	@return リテラル。"[^"]*"
**/
string parseLiteral(string str, ref int pos){
	assert(str.length > pos);
	int start_pos = pos;
	pos++;
	for(; pos<str.length; pos++){
		if(str[pos] == '"'){
			break;
		}
	}
	if(pos > str.length-1){
		pos = str.length-1;
	}
	return str[start_pos..pos+1];
}


/// 数字変数を解析する。
string parseIntegerVariable(string str, ref int pos){
	assert(str.length > pos + 1);
	int start_pos = pos;
	if(str[pos] == '%'){
		if(str[pos+1] == '%' || str[pos+1] == '?'){
			if(str.length-1 == pos+1){
				// 変数になってない。
				throw new Exception("input symbol or number");
			}
			pos++;
		}
		pos++;
	}
	// posで見るので戻り値は不要
	if(isSymbolCap(str[pos])){
		// シンボルを取得する。
		parseSymbol(str, pos);
	}
	else if(isdigit(str[pos])){
		// 数字列を取得する。
		parseNumber(str, pos);
	}
	else{
		// 変数になってない。
		throw new Exception("input symbol or number");
	}
	if(pos > str.length-1){
		pos = str.length-1;
	}
	return str[start_pos..pos+1];
}


/// 文字変数を解析する。
string parseStringVariable(string str, ref int pos){
	assert(str.length > pos + 1);
	int start_pos = pos;
	if(str[pos] == '$'){
		if(str[pos+1] == '%' || str[pos+1] == '?'){
			if(str.length-1 == pos+1){
				// 変数になってない。
				throw new Exception("input symbol or number");
			}
			pos++;
		}
		pos++;
	}
	// posで見るので戻り値は不要
	if(isSymbolCap(str[pos])){
		// シンボルを取得する。
		parseSymbol(str, pos);
	}
	else if(isdigit(str[pos])){
		// 数字列を取得する。
		parseNumber(str, pos);
	}
	else{
		// 変数になってない。
		throw new Exception("input symbol or number");
	}
	if(pos > str.length-1){
		pos = str.length-1;
	}
	return str[start_pos..pos+1];
}


/// 配列変数を解析する。
string parseArrayVariable(string str, ref int pos){
	assert(str.length > pos + 1);
	int start_pos = pos;
	if(str[pos] == '?'){
		if(str[pos+1] == '%'){
			if(str.length-1 == pos+1){
				// 変数になってない。
				throw new Exception("input symbol or number");
			}
			pos++;
		}
		pos++;
	}
	// posで見るので戻り値は不要
	if(isSymbolCap(str[pos])){
		// シンボルを取得する。
		parseSymbol(str, pos);
	}
	else if(isdigit(str[pos])){
		// 数字列を取得する。
		parseNumber(str, pos);
	}
	else{
		// 変数になってない。
		throw new Exception("input symbol or number");
	}
	if(pos > str.length-1){
		pos = str.length-1;
	}
	return str[start_pos..pos+1];
}


/// 数字列を取得する。
string parseNumber(string str, ref int pos){
	assert(str.length > pos);
	int start_pos = pos;
	pos++;
	for(; pos<str.length; pos++){
		if(!isdigit(str[pos])){
			pos--;
			break;
		}
	}
	if(pos > str.length-1){
		pos = str.length-1;
	}
	return str[start_pos..pos+1];
}


/// 16進数表記を取得する。
string parseXDigit(string str, ref int pos){
	int start_pos = pos;
	for(; pos<str.length; pos++){
		if(!isxdigit(str[pos])){
			pos--;
			break;
		}
	}
	if(pos > str.length-1){
		pos = str.length-1;
	}
	return str[start_pos..pos+1];
}


/// マークアップ文字列を取得する。
string parseMarkuppedString(string str, ref int pos, char start_mark, char end_mark){
	assert(str.length > pos+1);
	assert(str[pos] == start_mark);
	int start_pos = pos;
	pos++;
	for(; pos<str.length; pos++){
		if(str[pos] == end_mark){
			break;
		}
		// ネストは許容しない。
		if(str[pos] == start_mark){
			throw new Exception("invalid markup nest");
		}
	}
	if(pos > str.length-1){
		pos = str.length-1;
	}
	return str[start_pos..pos+1];
}


/// 文字列を取得する。
string parseString(string str, ref int pos){
	// UTF-8を前提とするため、ASCII文字があれば完了する。
	int start_pos = pos;
	for(; pos<str.length; pos++){
		if(isascii(str[pos])){
			pos--;
			break;
		}
	}
	if(pos > str.length-1){
		pos = str.length-1;
	}
	return str[start_pos..pos+1];
}

/// トークン群内に種別に適合するトークンがあるかどうか確認する。
bool isExistType(Token[] tokens, int type){
	assert(TokenType.min <= type && type <= TokenType.max);
	bool result = false;
	foreach(Token token; tokens){
		if(token.type == type){
			result = true;
		}
	}
	return result;
}


/// シンボルの先頭文字
bool isSymbolCap(char ch){
	if(isalpha(ch) || ch == '_'){
		return true;
	}
	return false;
}


/// シンボルの文字
bool isSymbolChar(char ch){
	if(isalnum(ch) || ch == '_'){
		return true;
	}
	return false;
}


/// テキストボタンかどうか
bool isTextButton(string str, int pos){
	bool ret = false;
	pos++;
	// 数字＋テキストかテキストのみ、その後に'>'が来る必要がある。
	// 数字なら数字以外まで動かしておく。
	if(isdigit(str[pos])){
		for(pos=pos+1; pos<str.length; pos++){
			if(!isdigit(str[pos])){
				break;
			}
		}
	}
	// 直後に非アスキーでその後に'>'が来る。
	for(; pos<str.length; pos++){
		if(isascii(str[pos])){
			// '>'が来ていればテキストボタンとみなせるはず。
			ret = (str[pos] == '>') ? true : false;
			break;
		}
	}
	return ret;
}

/// ルビ文字列かどうか
bool isRubyString(string str, int pos){
	bool ret = false;
	int start_pos = pos;
	pos++;
	if(isalnum(str[pos])){
		// 直後が英数字ならルビじゃない。
		return false;
	}
	if(str[pos] == '%' || str[pos] == '?'){
		// 直後が数字変数か配列変数ならルビじゃない。
		return false;
	}
	if(isascii(str[pos])){
		// 直後の英数字以外のアスキーは不正な構文であるため例外発生。
		throw new Exception("unexpected character '" ~ str[pos..pos+1] ~ "' just after '('.");
	}

	// 直後に非アスキーでその後に'/'、非アスキー、')'が来る。
	for(; pos<str.length; pos++){
		if(isascii(str[pos])){
			// '/'が来ているかどうか確認する。
			ret = (str[pos] == '/') ? true : false;
			break;
		}
	}
	if(!ret){
		// 非アスキーから'/'以外のアスキーは不正な構文であるため例外発生。
		throw new Exception("bad ruby format");
	}
	// '/'の後を調査
	ret = false;
	pos++;
	for(; pos<str.length; pos++){
		if(isascii(str[pos])){
			// ')'が来ていればルビとみなせるはず。
			ret = (str[pos] == ')') ? true : false;
			break;
		}
	}
	return ret;
}

/// 空白のトークンのみであるかどうか確認する。
bool isOnlySpace(Token[] tokens){
	bool ret = true;
	foreach(Token token; tokens){
		if(token.type != TokenType.SPACE){
			ret = false;
			break;
		}
	}
	return ret;
}


unittest
{
debug(uttoken) printf("parseToToken test...");

	// ルーチンテスト
	assert(isSymbolCap('a'));
	assert(isSymbolCap('n'));
	assert(isSymbolCap('z'));
	assert(isSymbolCap('A'));
	assert(isSymbolCap('N'));
	assert(isSymbolCap('Z'));
	assert(isSymbolCap('_'));
	assert(!isSymbolCap('0'));

	assert(isSymbolChar('a'));
	assert(isSymbolChar('n'));
	assert(isSymbolChar('z'));
	assert(isSymbolChar('A'));
	assert(isSymbolChar('N'));
	assert(isSymbolChar('Z'));
	assert(isSymbolChar('_'));
	assert(isSymbolChar('0'));
	assert(isSymbolChar('9'));

	assert(isRubyString("(薔薇/バラ)", 0));
	assert(!isRubyString("(27+1)", 0));
	
	Token[] tokens;
	// 単一語句解析テスト
	tokens = parseToTokens("適当な文字列");
	assert(tokens[0].value == "適当な文字列");
	assert(tokens[0].type == TokenType.STRING);
	
	tokens = parseToTokens(" ");
	assert(tokens[0].value == " ");
	assert(tokens[0].type == TokenType.SPACE);
	tokens = parseToTokens("    ");
	assert(tokens[0].value == "    ");
	assert(tokens[0].type == TokenType.SPACE);
	tokens = parseToTokens("\t");
	assert(tokens[0].value == "\t"); // 強制的に4つの空白に変換する?
	assert(tokens[0].type == TokenType.SPACE);

	tokens = parseToTokens(";comment");
	assert(tokens[0].value == ";comment");
	assert(tokens[0].type == TokenType.COMMENT);
	tokens = parseToTokens(";; comment $var %var \"\"");
	assert(tokens[0].value == ";; comment $var %var \"\"");
	assert(tokens[0].type == TokenType.COMMENT);

	tokens = parseToTokens("\"literal\"");
	assert(tokens[0].value == "\"literal\"");
	assert(tokens[0].type == TokenType.LITERAL);
	tokens = parseToTokens("\";; comment $var %var\"");
	assert(tokens[0].value == "\";; comment $var %var\"");
	assert(tokens[0].type == TokenType.LITERAL);

	tokens = parseToTokens("*define");
	assert(tokens[0].value == "*define");
	assert(tokens[0].type == TokenType.SPECIAL_LABEL);
	tokens = parseToTokens("*start");
	assert(tokens[0].value == "*start");
	assert(tokens[0].type == TokenType.SPECIAL_LABEL);
	tokens = parseToTokens("*customsel");
	assert(tokens[0].value == "*customsel");
	assert(tokens[0].type == TokenType.SPECIAL_LABEL);

	tokens = parseToTokens("*l");
	assert(tokens[0].value == "*l");
	assert(tokens[0].type == TokenType.LINETOP_LABEL);
	tokens = parseToTokens("*linetop");
	assert(tokens[0].value == "*linetop");
	assert(tokens[0].type == TokenType.LINETOP_LABEL);

	tokens = parseToTokens("%var");
	assert(tokens[0].value == "%var");
	assert(tokens[0].type == TokenType.NUMBER_VARIABLE);
	tokens = parseToTokens("%%var");
	assert(tokens[0].value == "%%var");
	assert(tokens[0].type == TokenType.NUMBER_VARIABLE);
	tokens = parseToTokens("%0");
	assert(tokens[0].value == "%0");
	assert(tokens[0].type == TokenType.NUMBER_VARIABLE);
	tokens = parseToTokens("%%0");
	assert(tokens[0].value == "%%0");
	assert(tokens[0].type == TokenType.NUMBER_VARIABLE);

	tokens = parseToTokens("$var");
	assert(tokens[0].value == "$var");
	assert(tokens[0].type == TokenType.STRING_VARIABLE);
	tokens = parseToTokens("$%var");
	assert(tokens[0].value == "$%var");
	assert(tokens[0].type == TokenType.STRING_VARIABLE);
	tokens = parseToTokens("$0");
	assert(tokens[0].value == "$0");
	assert(tokens[0].type == TokenType.STRING_VARIABLE);
	tokens = parseToTokens("$%0");
	assert(tokens[0].value == "$%0");
	assert(tokens[0].type == TokenType.STRING_VARIABLE);

	tokens = parseToTokens("?var");
	assert(tokens[0].value == "?var");
	assert(tokens[0].type == TokenType.ARRAY_VARIABLE);
	tokens = parseToTokens("?%var");
	assert(tokens[0].value == "?%var");
	assert(tokens[0].type == TokenType.ARRAY_VARIABLE);
	tokens = parseToTokens("?0");
	assert(tokens[0].value == "?0");
	assert(tokens[0].type == TokenType.ARRAY_VARIABLE);
	tokens = parseToTokens("?%0");
	assert(tokens[0].value == "?%0");
	assert(tokens[0].type == TokenType.ARRAY_VARIABLE);

	tokens = parseToTokens("#ffffff");
	assert(tokens[0].value == "#ffffff");
	assert(tokens[0].type == TokenType.COLOR);
	tokens = parseToTokens("#09af15");
	assert(tokens[0].value == "#09af15");
	assert(tokens[0].type == TokenType.COLOR);

	tokens = parseToTokens("1");
	assert(tokens[0].value == "1");
	assert(tokens[0].type == TokenType.NUMBER);
	tokens = parseToTokens("1234567890");
	assert(tokens[0].value == "1234567890");
	assert(tokens[0].type == TokenType.NUMBER);

	tokens = parseToTokens("a");
	assert(tokens[0].value == "a");
	assert(tokens[0].type == TokenType.SYMBOL);
	tokens = parseToTokens("_aA0");
	assert(tokens[0].value == "_aA0");
	assert(tokens[0].type == TokenType.SYMBOL);

	tokens = parseToTokens("!s");
	assert(tokens[0].value == "!s");
	assert(tokens[0].type == TokenType.SPECIAL_CHARACTER);

	tokens = parseToTokens("/");
	assert(tokens[0].value == "/");
	assert(tokens[0].type == TokenType.SPECIAL_CHARACTER);

	tokens = parseToTokens("/2");
	assert(tokens[0].value == "/");
	assert(tokens[0].type == TokenType.OPERATOR);

	tokens = parseToTokens("+");
	assert(tokens[0].value == "+");
	assert(tokens[0].type == TokenType.OPERATOR);

	tokens = parseToTokens("-");
	assert(tokens[0].value == "-");
	assert(tokens[0].type == TokenType.OPERATOR);

	tokens = parseToTokens("(瑠微/ルビ)");
	assert(tokens[0].value == "(瑠微/ルビ)");
	assert(tokens[0].type == TokenType.RUBY);

	tokens = parseToTokens("[タグ]");
	assert(tokens[0].value == "[タグ]");
	assert(tokens[0].type == TokenType.TAG);

	tokens = parseToTokens("<テキストボタン>");
	assert(tokens[0].value == "<テキストボタン>");
	assert(tokens[0].type == TokenType.TEXT_BUTTON);
	tokens = parseToTokens("<5テキストボタン>");
	assert(tokens[0].value == "<5テキストボタン>");
	assert(tokens[0].type == TokenType.TEXT_BUTTON);

	// 複合語句解析テスト
	tokens = parseToTokens("[戸橋]「よう、$F111！_」");
	assert(tokens.length == 6);
	assert(tokens[0].value == "[戸橋]");
	assert(tokens[0].type == TokenType.TAG);
	assert(tokens[1].value == "「よう、");
	assert(tokens[1].type == TokenType.STRING);
	assert(tokens[2].value == "$F111");
	assert(tokens[2].type == TokenType.STRING_VARIABLE);
	assert(tokens[3].value == "！");
	assert(tokens[3].type == TokenType.STRING);
	assert(tokens[4].value == "_");
	assert(tokens[4].type == TokenType.SPECIAL_CHARACTER);
	assert(tokens[5].value == "」");
	assert(tokens[5].type == TokenType.STRING);

	tokens = parseToTokens("if %0=0 goto *text_end");
	/*
	printf("\n%d\n", tokens.length);
	foreach(Token token; tokens){
		token.dump();
		printf("\n");
	}
	*/
	assert(tokens.length == 9);
	assert(tokens[0].value == "if");
	assert(tokens[0].type == TokenType.SYMBOL);
	assert(tokens[1].value == " ");
	assert(tokens[1].type == TokenType.SPACE);
	assert(tokens[2].value == "%0");
	assert(tokens[2].type == TokenType.NUMBER_VARIABLE);
	assert(tokens[3].value == "=");
	assert(tokens[3].type == TokenType.OPERATOR);
	assert(tokens[4].value == "0");
	assert(tokens[4].type == TokenType.NUMBER);
	assert(tokens[5].value == " ");
	assert(tokens[5].type == TokenType.SPACE);
	assert(tokens[6].value == "goto");
	assert(tokens[6].type == TokenType.SYMBOL);
	assert(tokens[7].value == " ");
	assert(tokens[7].type == TokenType.SPACE);
	assert(tokens[8].value == "*text_end");
	assert(tokens[8].type == TokenType.GENERIC_LABEL);

	assert(isTextButton("<テキストボタン>", 0));
	assert(isTextButton("<10テキストボタン>", 0));
	assert(!isTextButton("if %0<10 return", 5));
	assert(!isTextButton("if %0<10 && %1>9 return", 5));

	tokens = parseToTokens("if %flag_len<1 mesbox \"sflag: invalid length\",\"ASSERT!\":end");
	assert(tokens.length == 13);
	assert(tokens[0].value == "if");
	assert(tokens[0].type == TokenType.SYMBOL);
	assert(tokens[1].value == " ");
	assert(tokens[1].type == TokenType.SPACE);
	assert(tokens[2].value == "%flag_len");
	assert(tokens[2].type == TokenType.NUMBER_VARIABLE);
	assert(tokens[3].value == "<");
	assert(tokens[3].type == TokenType.OPERATOR);
	assert(tokens[4].value == "1");
	assert(tokens[4].type == TokenType.NUMBER);
	assert(tokens[5].value == " ");
	assert(tokens[5].type == TokenType.SPACE);
	assert(tokens[6].value == "mesbox");
	assert(tokens[6].type == TokenType.SYMBOL);
	assert(tokens[7].value == " ");
	assert(tokens[7].type == TokenType.SPACE);
	assert(tokens[8].value == "\"sflag: invalid length\"");
	assert(tokens[8].type == TokenType.LITERAL);
	assert(tokens[9].value == ",");
	assert(tokens[9].type == TokenType.ASCII);
	assert(tokens[10].value == "\"ASSERT!\"");
	assert(tokens[10].type == TokenType.LITERAL);
	assert(tokens[11].value == ":");
	assert(tokens[11].type == TokenType.OPERATOR);
	assert(tokens[12].value == "end");
	assert(tokens[12].type == TokenType.SYMBOL);
	
	// !=の比較考慮漏れ
	tokens = parseToTokens("if $ut_testcaselist!=\"\" add $%ut_testfixture,TESTCASE_SEPARATOR");
	assert(tokens.length == 11);
	assert(tokens[0].value == "if");
	assert(tokens[0].type == TokenType.SYMBOL);
	assert(tokens[1].value == " ");
	assert(tokens[1].type == TokenType.SPACE);
	assert(tokens[2].value == "$ut_testcaselist");
	assert(tokens[2].type == TokenType.STRING_VARIABLE);
	assert(tokens[3].value == "!=");
	assert(tokens[3].type == TokenType.OPERATOR);
	assert(tokens[4].value == "\"\"");
	assert(tokens[4].type == TokenType.LITERAL);
	assert(tokens[5].value == " ");
	assert(tokens[5].type == TokenType.SPACE);
	assert(tokens[6].value == "add");
	assert(tokens[6].type == TokenType.SYMBOL);
	assert(tokens[7].value == " ");
	assert(tokens[7].type == TokenType.SPACE);
	assert(tokens[8].value == "$%ut_testfixture");
	assert(tokens[8].type == TokenType.STRING_VARIABLE);
	assert(tokens[9].value == ",");
	assert(tokens[9].type == TokenType.ASCII);
	assert(tokens[10].value == "TESTCASE_SEPARATOR");
	assert(tokens[10].type == TokenType.SYMBOL);

	// 演算用の小括弧考慮漏れ
	tokens = parseToTokens("(27+1)");
	assert(tokens.length == 5);
	assert(tokens[0].value == "(");
	assert(tokens[0].type == TokenType.PARENTHESIS_BEGIN);
	assert(tokens[1].value == "27");
	assert(tokens[1].type == TokenType.NUMBER);
	assert(tokens[2].value == "+");
	assert(tokens[2].type == TokenType.OPERATOR);
	assert(tokens[3].value == "1");
	assert(tokens[3].type == TokenType.NUMBER);
	assert(tokens[4].value == ")");
	assert(tokens[4].type == TokenType.PARENTHESIS_END);

	// 配列変数の要素指定時の大括弧ネスト考慮漏れ
	tokens = parseToTokens("?array[?array[0]]");
	assert(tokens.length == 7);
	assert(tokens[0].value == "?array");
	assert(tokens[0].type == TokenType.ARRAY_VARIABLE);
	assert(tokens[1].value == "[");
	assert(tokens[1].type == TokenType.BRACKET_BEGIN);
	assert(tokens[2].value == "?array");
	assert(tokens[2].type == TokenType.ARRAY_VARIABLE);
	assert(tokens[3].value == "[");
	assert(tokens[3].type == TokenType.BRACKET_BEGIN);
	assert(tokens[4].value == "0");
	assert(tokens[4].type == TokenType.NUMBER);
	assert(tokens[5].value == "]");
	assert(tokens[5].type == TokenType.BRACKET_END);
	assert(tokens[6].value == "]");
	assert(tokens[6].type == TokenType.BRACKET_END);

	// 無名ラベルが利用できることを発見
	tokens = parseToTokens("*");
	assert(tokens[0].value == "*", tokens[0].value);
	assert(tokens[0].type == TokenType.LINETOP_LABEL);
	tokens = parseToTokens("gosub *");
	assert(tokens[0].value == "gosub");
	assert(tokens[0].type == TokenType.SYMBOL);
	assert(tokens[1].value == " ");
	assert(tokens[1].type == TokenType.SPACE);
	assert(tokens[2].value == "*");
	assert(tokens[2].type == TokenType.GENERIC_LABEL);

	// 配列変数を利用した変数番号指定の考慮漏れ
	tokens = parseToTokens("$?0[0]");
	assert(tokens[0].value == "$?0");
	assert(tokens[0].type == TokenType.STRING_VARIABLE);
	assert(tokens[1].value == "[");
	assert(tokens[1].type == TokenType.BRACKET_BEGIN);
	assert(tokens[2].value == "0");
	assert(tokens[2].type == TokenType.NUMBER);
	assert(tokens[3].value == "]");
	assert(tokens[3].type == TokenType.BRACKET_END);
	tokens = parseToTokens("%?0[0]");
	assert(tokens[0].value == "%?0");
	assert(tokens[0].type == TokenType.NUMBER_VARIABLE);
	assert(tokens[1].value == "[");
	assert(tokens[1].type == TokenType.BRACKET_BEGIN);
	assert(tokens[2].value == "0");
	assert(tokens[2].type == TokenType.NUMBER);
	assert(tokens[3].value == "]");
	assert(tokens[3].type == TokenType.BRACKET_END);

debug(uttoken) printf("passed\n");
}

debug(uttoken)
{
	int main(){
		return 0;
	}
}
