import std.xml;
import std.stdio;
import std.string;
import std.conv;
import std.getopt;

import xnovel;
import nscr.token;
import nscr.charset;


int main(string[] args){
    if(args.length <= 1) return usage();
    string output;
    bool print;
    getopt(
        args,
        "output|o", &output,
        "print|p", &print
    );
    if(args.length <= 1) return usage();
    string execdir = std.path.dirname(std.path.rel2abs(args[0]));
    string input = args[1];
    if(output.length == 0) output = std.path.join(execdir, "main.html");
    if(!std.path.isabs(input)) input = std.path.join(execdir, input);
    if(!std.file.exists(input)){
        writeln(input, " Not Found.");
        return 1;
    }
    string nscript = cast(string)std.file.read(input);
    Token[][] lines = parseToTokenLines(splitlines(fromNScrCharset(nscript)));
    auto builder = new XNovelBuilder();
    bool definition = true;
    bool contents = false;
    Command[] prevCommands;
    auto analyzeOnDefinition = (Token[] tokens){
        auto tokenStream = new TokenStream(tokens);
        if(!tokenStream.skipToAffectable()) return;
        auto token = tokenStream.next();
        // Definition not needed.
        if(token.type == TokenType.SYMBOL){
            switch(token.value){
            case "game":
                definition = false;
                break;
            case "stralias":
                straliasTokensToBuild(builder, tokenStream);
                break;
            default:
                // Unsupported command.
                break;
            }
        }
    };
    auto analyzeOnExecution = (Token[] tokens){
        auto tokenStream = new TokenStream(tokens);
        if(!tokenStream.skipToAffectable()) return;
        auto token = tokenStream.next();
        if(tokens[0].type == TokenType.SPECIAL_LABEL && tokens[0].value == "*start")
            contents = true;
    };
    auto analyzeOnContents = (Token[] tokens){
        // set nav/h1
        auto commands = tokensToCommands(tokens);
        if(typeid(commands[0]) == typeid(Nav) && isTextCommands(prevCommands)){
            auto nav = cast(Nav)commands[0];
            //string[string];
            nav._title = H1.create(null, toText(prevCommands[0..$-1]));
            builder.back(prevCommands.length);
        }
        builder.add(commands);
        prevCommands = commands;
    };
    int lineno = 1;
    // analyze
    foreach(tokens; lines){
        try{
            if(tokens.length == 0) {
                if(contents) builder.add(new Blank);
                continue;
            }
            if(definition){
                analyzeOnDefinition(tokens);
            }
            else{
                if(!contents){
                    analyzeOnExecution(tokens);
                }
                else{
                    analyzeOnContents(tokens);
                }
            }
        }
        catch(Exception e){
            writeln("line " ~ to!string(lineno) ~ ": " ~ toNScrCharset(e.msg));
            return 10;
        }
        debug{
            foreach(token; tokens){
                write(toNScrCharset(token.value));
            }
            writeln("");
        }
        lineno++;
    }
    XNovel xn = builder.create();
    string xnxml = xn.toXML();
    if(print){
        writeln(toNScrCharset(xnxml));
    }
    else{
        std.file.write(output, xnxml);
    }
    return 0;
}

int usage(){
    writeln("
Usage:
  ns2xn [-option] <input file>
  
  input file                    input file name like 00.txt(Shift_JIS)
  --output=<output file>, -o    output file name like 00.html(UTF-8)
                                default main.html
  --print, -p                   only print, not output to file
");
    return 0;
}

string toText(Command[] commands){
    string text;
    foreach(command; commands){
        text ~= join(command.pretty(0), "");
    }
    return text;
}

bool isTextCommands(Command[] commands){
    int index = 0;
    int last = commands.length-1; // The last is Br.
    while(index < last){
        auto type = typeid(commands[index++]);
        if(type != typeid(xnovel.Text) && type != typeid(Ruby))
            return false;
    }
    return true;
}

Command[] tokensToCommands(Token[] tokens){
    Command[] commands;
    TokenStream tokenStream = new TokenStream(tokens);
    try {
        while(tokenStream.hasNext()){
            Token* token = tokenStream.next();
            switch(token.type){
            case TokenType.SYMBOL:
                switch(token.value){
                case "bg":
                    commands ~= bgTokensToCommand(tokenStream);
                    break;
                case "ld":
                    commands ~= ldTokensToCommand(tokenStream);
                    break;
                case "bgm":
                    commands ~= bgmTokensToCommand(tokenStream);
                    break;
                case "select":
                    commands ~= selectTokensToCommand(tokenStream);
                    break;
                case "bgmstop":
                    commands ~= new XNStopbgm;
                    break;
                case "end":
                    commands ~= new XNClose;
                    break;
                case "movie":
                    commands ~= movieTokensToCommand(tokenStream);
                    break;
                case "wait":
                    commands ~= waitTokensToCommand(tokenStream);
                    break;
                case "mov":
                    commands ~= movTokensToCommand(tokenStream);
                    break;
                case "if":
                    commands ~= ifTokensToCommand(tokenStream);
                    break;
                case "input":
                    commands ~= inputTokensToCommand(tokenStream);
                    break;
                case "wave":
                case "waveloop":
                    commands ~= waveTokensToCommand(tokenStream);
                    break;
                case "dwave":
                case "dwaveloop":
                    commands ~= dwaveTokensToCommand(tokenStream);
                    break;
                case "stop":
                    commands ~= new XNStopAllse;
                    break;
                default:
                    // ignore
                }
                break;
            
            case TokenType.STRING:
                commands ~= new xnovel.Text(token.value);
                tokenStream.skipSpace();
                if(!tokenStream.hasNext() || tokenStream.isNextType(TokenType.COMMENT)){
                    commands ~= new Br;
                }
                break;
            
            case TokenType.RUBY:
                string[] rbrt = std.string.split(token.value[1..$-1], "/");
                commands ~= Ruby.create(rbrt[0], rbrt[1]);
                if(!tokenStream.hasNext() || tokenStream.isNextType(TokenType.COMMENT)){
                    commands ~= new Br;
                }
                break;
            
            case TokenType.COMMENT:
                commands ~= new XNNop(token.value[1..$]);
                break;
            
            case TokenType.STRING_VARIABLE:
            case TokenType.NUMBER_VARIABLE:
                commands ~= new xnovel.Text("#{" ~ var2name(token.value) ~ "}");
                if(!tokenStream.hasNext() || tokenStream.isNextType(TokenType.COMMENT)){
                    commands ~= new Br;
                }
                break;
            
            case TokenType.LINETOP_LABEL:
                commands ~= Br.create(["id":token.value[1..$]]);
                break;
            default:
                token.dump();
                writeln("");
                break;
            }
        }
    }
    catch(Exception e){
        debug writeln(toNScrCharset(e.toString()));
        throw e;
    }
    return commands;
}

void straliasTokensToBuild(ref XNovelBuilder builder, TokenStream tokenStream){
    if(!tokenStream.skipToFirstParam()) throw new Exception("最初のパラメータがありません");
    Token* token = tokenStream.next();
    if(token.type != TokenType.SYMBOL) throw new Exception("名前がありません");
    auto name = tolower(token.value);
    if(!tokenStream.skipToNextParam()) throw new Exception("次のパラメータがありません");
    token = tokenStream.next();
    if(token.type != TokenType.LITERAL) throw new Exception("文字列がありません");
    auto value = dequote(token.value);
    switch(name){
    case "xn_head_title":
        builder.setTitle(value);
        break;
    case "xn_body_title":
        builder.setBodyTitle(value);
        break;
    case "xn_direction":
    case "xn_text_rotation":
    case "xn_support":
    case "xn_curl":
        builder.setSettings(name, value);
        break;
    default:
        auto suffix = chompPrefix(name, "xn_body_");
        if(suffix != "xn_body_"){
            builder.setBodyInfo(suffix, value);
        }
    }
}

class TokenStream {
    private Token[] _tokens;
    private int _index;
    
    this(Token[] tokens){
        _tokens = tokens;
        _index = 0;
    }
    
    bool hasNext(){
        return _index < _tokens.length;
    }
    
    Token* next(){
        return &_tokens[_index++];
    }
    
    void skipSpace(){
        while(_index < _tokens.length && _tokens[_index].type == TokenType.SPACE)
            _index++;
    }
    
    bool skipToAffectable(){
        skipSpace();
        if(!hasNext()) return false;
        return _tokens[_index].type != TokenType.COMMENT;
    }
    
    bool skipToFirstParam(){
        if(!skipToAffectable()) return false;
        return isAcceptableAsParam(_tokens[_index].type);
    }
    
    bool skipToNextParam(){
        skipSpace();
        if(!hasNext()) return false;
        if(_tokens[_index].type == TokenType.ASCII && _tokens[_index].value == ","){
            _index++;
            skipSpace();
            if(!hasNext()) return false;
            return isAcceptableAsParam(_tokens[_index].type);
        }
        return false;
    }
    
    bool isNextType(TokenType type){
        return _tokens[_index].type == type;
    }
    
    bool isPreviousValue(string value){
        return _tokens[_index - 1].value == value;
    }
}

unittest {
    writeln("Test Parser");
    string s = cast(string)std.file.read("main.html");
    check(s);
    auto xn = new XNovel(s);
    assert(xn._head._title == "famibee XNovel紹介");
    assert(xn._body._title == "XNovel紹介");
    Contents contents = xn._body._contents;
    Command[] commands = contents._commands;
    assert(commands.length == 2, to!string(commands.length));
    assert(commands[1].classinfo.name == "xnovel.XNStopbgm", commands[1].classinfo.name);
    P p = cast(P)(commands[0]);
    Command[] children = p._children;
    assert(children.length > 0);
    int order = 0;
    assert(typeid(children[order++]) == typeid(Img), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Fg), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Audio), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(xnovel.Text), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Br), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Nav), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Br), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Ruby), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(xnovel.Text), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Br), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(XNClose), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(XNWait), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Video), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(XNLet), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(XNTrace), children[order].classinfo.name);
    Img img = cast(Img)(children[0]);
    assert(img.id == "bg");
    assert(img.src == "bg/machi.jpg");
    Fg fg = cast(Fg)(children[1]);
    assert(fg.id == "0");
    assert(fg.src == "fg/usagi01.png");
    assert(fg.tate_pos == 2);
    assert(fg.yoko_pos == 8);
    Audio audio = cast(Audio)(children[2]);
    assert(audio.src == "bgm/music.mp3");
    assert(audio.loop);
    assert(!audio.autoplay);
    auto text = cast(xnovel.Text)(children[3]);
    assert(text.text == "うさぎちゃん「このアプリはどんなアプリなの？」", text.text);
}

unittest
{
    writeln("Test Builder");
    auto builder = new XNovelBuilder();
    auto xn = builder
                .setTitle("Builder Test")
                .setBodyTitle("Title")
                .setBodyInfo("author", "Author")
                .add(Img.create(["id":"bg", "src":"bg/hoge.jpg"]))
                .add(new xnovel.Text("ほげ"))
                .create();
    check(xn.toXML());
}

unittest {
    assert(dequote("\"hoge\"") == "hoge");
    assert(millis2seconds("3000") == "3", millis2seconds("3000"));
    assert(millis2seconds("500") == "0.5", millis2seconds("500"));
    assert(millis2seconds("1") == "0.001", millis2seconds("1"));
    assert(var2name("%0") == "save:i0", var2name("%0"));
    assert(var2name("$0") == "save:s0", var2name("$0"));
    assert(var2name("%200") == "sys:i200", var2name("%200"));
    assert(var2name("$200") == "sys:s200", var2name("$200"));
    assert(lit2squote("\"<ほげ>&\"") == "'&lt;ほげ&gt;&amp;'", lit2squote("\"<ほげ>&\""));
    
    Token[] tokens;
    Command[] commands;
    
    tokens = parseToTokens("bg \"bg/hoge.jpg\",1");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Img));
    auto img = cast(Img)commands[0];
    assert(img.id == "bg");
    assert(img.src == "bg/hoge.jpg", img.src);

    tokens = parseToTokens("ld l,\"fg/hoge.png\",1");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Fg));
    auto fg = cast(Fg)commands[0];
    assert(fg.id == "0");
    assert(fg.src == "fg/hoge.png", fg.src);
    assert(fg.tate_pos == 1, to!string(fg.tate_pos));
    assert(fg.yoko_pos == 7, to!string(fg.yoko_pos));

    Audio audio;
    tokens = parseToTokens("bgm \"bgm/hoge.mp3\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Audio));
    audio = cast(Audio)commands[0];
    assert(audio.src == "bgm/hoge.mp3", audio.src);
    assert(audio.loop);
    assert(audio.autoplay);
    assert(audio.id == "bgm", audio.id);
    
    xnovel.Text text;
    tokens = parseToTokens("ほげほげ");
    commands = tokensToCommands(tokens);
    assert(commands.length == 2, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(xnovel.Text));
    text = cast(xnovel.Text)commands[0];
    assert(text.text == "ほげほげ", text.text);
    assert(typeid(commands[1]) == typeid(Br));
    
//    tokens = parseToTokens("選択してください");
    tokens = parseToTokens("select \"先頭へ\",*sel0,\"中間へ\",*sel1,\"後半へ\",*sel2,\"最後へ\",*sel3");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Nav));
    auto nav = cast(Nav)commands[0];
    // TODO Test nav > h1
    assert(nav._links.length == 4, to!string(nav._links.length));
    assert(nav._links[0].text == "先頭へ");
    assert(nav._links[0].href == "#sel0");
    
    tokens = parseToTokens("*sel0");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Br));
    auto br = cast(Br)commands[0];
    assert(br.id == "sel0", br.id);
    
    Ruby ruby;
    tokens = parseToTokens("(先頭/せんとう)です。");
    commands = tokensToCommands(tokens);
    assert(commands.length == 3, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Ruby));
    ruby = cast(Ruby)commands[0];
    assert(ruby._rb == "先頭", ruby._rb);
    assert(ruby._rt == "せんとう", ruby._rt);
    assert(ruby._rps[0] == "(", ruby._rps[0]);
    assert(ruby._rps[1] == ")", ruby._rps[1]);
    assert(typeid(commands[1]) == typeid(xnovel.Text));
    text = cast(xnovel.Text)commands[1];
    assert(text.text == "です。", text.text);
    assert(typeid(commands[2]) == typeid(Br));
    
    tokens = parseToTokens("bgmstop");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNStopbgm));
    
    tokens = parseToTokens("end");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNClose));

    tokens = parseToTokens("movie \"video/hoge.mp4\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Video));
    auto video = cast(Video)commands[0];
    assert(video.src == "video/hoge.mp4", audio.src);

    tokens = parseToTokens("wait 3000");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNWait));
    auto xn_wait = cast(XNWait)commands[0];
    assert(xn_wait.time == 3, to!string(xn_wait.time));
    
    // Test breaking if terminated by ruby.
    tokens = parseToTokens("これで(最後/さいご)");
    commands = tokensToCommands(tokens);
    assert(commands.length == 3, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(xnovel.Text));
    text = cast(xnovel.Text)commands[0];
    assert(text.text == "これで", text.text);
    assert(typeid(commands[1]) == typeid(Ruby));
    ruby = cast(Ruby)commands[1];
    assert(ruby._rb == "最後", ruby._rb);
    assert(ruby._rt == "さいご", ruby._rt);
    assert(ruby._rps[0] == "(", ruby._rps[0]);
    assert(ruby._rps[1] == ")", ruby._rps[1]);
    assert(typeid(commands[2]) == typeid(Br));
    
    XNLet xn_let;
    // $VAR, STR
    tokens = parseToTokens("mov $0,\"local\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNLet));
    xn_let = cast(XNLet)commands[0];
    assert(xn_let.name == "save:s0", xn_let.name);
    assert(xn_let.text == "'local'", xn_let.text);
    assert(!xn_let.calc);
    // %VAR, NUM
    tokens = parseToTokens("mov %0,100");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNLet));
    xn_let = cast(XNLet)commands[0];
    assert(xn_let.name == "save:i0", xn_let.name);
    assert(xn_let.text == "100", xn_let.text);
    assert(xn_let.calc);
    
    tokens = parseToTokens("if %0 > 99 goto *q_easy");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(A));
    auto a = cast(A)commands[0];
    assert(a.href == "#q_easy", a.href);
    assert(a.cond == "save:i0 &gt; 99", a.cond);
    
    tokens = parseToTokens("; 後ろ2つの数値は無視されます。");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNNop));
    auto xn_nop = cast(XNNop)commands[0];
    assert(xn_nop.text == " 後ろ2つの数値は無視されます。", xn_nop.text);
    
    tokens = parseToTokens("input $0,\"文字列を入力してください\",\"デフォルト\",24,0");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Input));
    auto input = cast(Input)commands[0];
    assert(input.title == "文字列を入力してください", input.title);
    assert(input.name == "save:s0", input.name);
    assert(input.placeholder == "デフォルト", input.placeholder);
    assert(input.maxlength == 24, to!string(input.maxlength));

    tokens = parseToTokens("変数の中身は$0");
    commands = tokensToCommands(tokens);
    assert(commands.length == 3, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(xnovel.Text));
    text = cast(xnovel.Text)commands[0];
    assert(text.text == "変数の中身は", text.text);
    assert(typeid(commands[1]) == typeid(xnovel.Text));
    text = cast(xnovel.Text)commands[1];
    assert(text.text == "#{save:s0}", text.text);
    assert(typeid(commands[2]) == typeid(Br));
    
    // wave
    tokens = parseToTokens("wave \"se/hoge.wav\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Audio));
    audio = cast(Audio)commands[0];
    assert(audio.src == "se/hoge.wav", audio.src);
    assert(!audio.loop);
    assert(audio.autoplay);
    assert(audio.id is null, audio.id);
    // waveloop
    tokens = parseToTokens("waveloop \"se/hoge.wav\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Audio));
    audio = cast(Audio)commands[0];
    assert(audio.src == "se/hoge.wav", audio.src);
    assert(audio.loop);
    assert(audio.autoplay);
    assert(audio.id is null, audio.id);
    // dwave
    tokens = parseToTokens("dwave 0,\"se/fuga.wav\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Audio));
    audio = cast(Audio)commands[0];
    assert(audio.src == "se/fuga.wav", audio.src);
    assert(!audio.loop);
    assert(audio.autoplay);
    assert(audio.id == "0", audio.id);
    // dwaveloop
    tokens = parseToTokens("dwaveloop 1,\"se/fuga.wav\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Audio));
    audio = cast(Audio)commands[0];
    assert(audio.src == "se/fuga.wav", audio.src);
    assert(audio.loop);
    assert(audio.autoplay);
    assert(audio.id == "1", audio.id);
    
    tokens = parseToTokens("stop");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNStopAllse));
    
}

string dequote(string str) {
    if(str.length > 1 && str[0] == '"' && str[$-1] == '"')
        return str[1..$-1];
    return str;
}

string millis2seconds(string time){
    int milli = to!int(time);
    string sec = format("%1.3f", (milli / 1000.0f));
    auto m = std.regexp.search(sec, "0+$");
    if(!m) return sec;
    string zero = m.match(0);
    if(zero.length == 3) return sec[0..1];
    return sec[0..$-zero.length];
}

string var2name(string var){
    // エイリアスは受け付けない。
    string infix = var[0] == '$' ? "s" : "i";
    string suffix = var[1..$];
    int varnum = to!int(suffix);
    return (varnum < 200 ? "save:" : "sys:") ~ infix ~ suffix;
}

string lit2squote(string lit){
    string result = std.xml.encode(dequote(lit));
    return "'" ~ result ~ "'";
}

Command bgTokensToCommand(TokenStream tokenStream) {
    // bg STR,NUM
    tokenStream.skipSpace();
    Token* token = tokenStream.next();
    if(token.type != TokenType.LITERAL) throw new Exception("文字列以外は対応していません");
    string src = dequote(token.value);
    
    if(!tokenStream.skipToNextParam()) throw new Exception("意図しないコードがあります");
    tokenStream.next(); // NUM
    
    if(tokenStream.skipToNextParam()){
        // bg STR,NUM,NUM
        token = tokenStream.next(); // NUM
        if(token.type != TokenType.NUMBER) throw new Exception("文字列以外は対応していません");
        return Img.create(["id":"bg", "src":src, "time":millis2seconds(token.value)]);
    }
    return Img.create(["id":"bg", "src":src]);
}

Command ldTokensToCommand(TokenStream tokenStream) {
    // ld NAME,STR,NUM
    tokenStream.skipSpace();
    Token* token = tokenStream.next();
    if(token.type != TokenType.SYMBOL) throw new Exception("位置が指定されていません");
    string id;
    string tate_pos;
    string yoko_pos;
    switch(token.value){
    case "l":
        id = "0";
        tate_pos = "1";
        yoko_pos = "7";
        break;
    case "c":
        id = "1";
        tate_pos = "2";
        yoko_pos = "8";
        break;
    case "r":
        id = "2";
        tate_pos = "3";
        yoko_pos = "9";
        break;
    default:
        throw new Exception("不正な位置です:" ~ token.value);
    }

    if(!tokenStream.skipToNextParam()) throw new Exception("意図しないコードがあります");
    token = tokenStream.next();
    if(token.type != TokenType.LITERAL) throw new Exception("文字列以外は対応していません");
    string src = dequote(token.value);
    
    if(!tokenStream.skipToNextParam()) throw new Exception("意図しないコードがあります");
    tokenStream.next(); // NUM
    
    if(tokenStream.skipToNextParam()){
        // ld NAME,STR,NUM,NUM
        token = tokenStream.next(); // NUM
        if(token.type != TokenType.NUMBER) throw new Exception("文字列以外は対応していません");
        return Fg.create(["id":id, "src":src, "tate_pos":tate_pos, "yoko_pos":yoko_pos, "time":millis2seconds(token.value)]);
    }
    return Fg.create(["id":id, "src":src, "tate_pos":tate_pos, "yoko_pos":yoko_pos]);
}

Command bgmTokensToCommand(TokenStream tokenStream) {
    // bgm STR
    tokenStream.skipSpace();
    Token* token = tokenStream.next();
    if(token.type != TokenType.LITERAL) throw new Exception("文字列以外は対応していません");
    string src = dequote(token.value);
    return Audio.create(["id":"bgm", "src":src, "loop":XN_TRUE, "autoplay":XN_TRUE]);
}

Command selectTokensToCommand(TokenStream tokenStream) {
    // select STR,LABEL,...
    auto nav = new Nav();
    tokenStream.skipSpace();
    Token* token;
    do{
        token = tokenStream.next();
        if(token.type != TokenType.LITERAL) throw new Exception("文字列以外は対応していません");
        string text = dequote(token.value);
        
        if(!tokenStream.skipToNextParam()) throw new Exception("意図しないコードがあります");
        token = tokenStream.next();
        if(token.type != TokenType.GENERIC_LABEL) throw new Exception("ラベル以外は対応していません");
        string href = "#" ~ token.value[1..$];
        
        nav._links ~= A.create(["href":href], text);
    }while(tokenStream.skipToNextParam());
    return nav;
}

Command movieTokensToCommand(TokenStream tokenStream) {
    // movie STR
    tokenStream.skipSpace();
    Token* token = tokenStream.next();
    if(token.type != TokenType.LITERAL) throw new Exception("文字列以外は対応していません");
    string src = dequote(token.value);
    return Video.create(["src":src]);
}

Command waitTokensToCommand(TokenStream tokenStream) {
    // wait NUM
    tokenStream.skipSpace();
    Token* token = tokenStream.next();
    if(token.type != TokenType.NUMBER) throw new Exception("数値以外は対応していません");
    return XNWait.create(["time":millis2seconds(token.value)]);
}

Command movTokensToCommand(TokenStream tokenStream) {
    // mov VAR,STR|NUM
    tokenStream.skipSpace();
    Token* token = tokenStream.next();
    string name;
    string value;
    string calc;
    if(token.type == TokenType.NUMBER_VARIABLE) {
        name = var2name(token.value);
        calc = XN_TRUE;
        if(!tokenStream.skipToNextParam()) throw new Exception("意図しないコードがあります");
        // 計算式と見て、modの変換以外はすべて受け付ける。
        calc_loop:
        do{
            token = tokenStream.next();
            switch(token.type){
            case TokenType.NUMBER:
            case TokenType.PARENTHESIS_BEGIN:
            case TokenType.PARENTHESIS_END:
            case TokenType.SPACE:
                value ~= token.value;
                break;
            case TokenType.OPERATOR:
                if(!inPattern(token.value[0], "-+*/")) throw new Exception("意図しないコードがあります");
                value ~= token.value;
                break;
            case TokenType.SYMBOL:
                if(token.value != "mod") throw new Exception("数値以外は対応していません");
                value ~= "%";
                break;
            case TokenType.COMMENT:
                break calc_loop;
            default:
                token.dump();
                throw new Exception("意図しないコードがあります");
            }
        }while(tokenStream.hasNext());
    }
    else if(token.type == TokenType.STRING_VARIABLE) {
        name = var2name(token.value);
        calc = XN_FALSE;
        if(!tokenStream.skipToNextParam()) throw new Exception("意図しないコードがあります");
        // 文字列結合と見て、modの変換以外はすべて受け付ける。
        cat_loop:
        do{
            token = tokenStream.next();
            switch(token.type){
            case TokenType.LITERAL:
                value ~= lit2squote(token.value);
                break;
            case TokenType.SPACE:
                value ~= token.value;
                break;
            case TokenType.OPERATOR:
                // XNovel側で結合されるかはなぞ。
                if(token.value != "+") throw new Exception("意図しないコードがあります");
                calc = XN_TRUE;
                value ~= token.value;
                break;
            case TokenType.COMMENT:
                break cat_loop;
            default:
                token.dump();
                throw new Exception("意図しないコードがあります");
            }
        }while(tokenStream.hasNext());
    }
    else {
        throw new Exception("意図しないコードがあります");
    }
    return XNLet.create(["name":name, "text":value, "calc":calc]);
}

Command ifTokensToCommand(TokenStream tokenStream) {
    // if CONDITION COMMAND
    // COMMAND only "goto LABEL"
    tokenStream.skipSpace();
    Token* token;
    string cond;
    // COMMANDまでは条件句
    cond_loop:
    do{
        token = tokenStream.next();
        switch(token.type){
        case TokenType.NUMBER_VARIABLE:
        case TokenType.STRING_VARIABLE:
            cond ~= var2name(token.value);
            break;
        case TokenType.OPERATOR:
        case TokenType.ASCII:
            // TODO パターン検証
            cond ~= std.xml.encode(token.value);
            break;
        case TokenType.NUMBER:
        case TokenType.PARENTHESIS_BEGIN:
        case TokenType.PARENTHESIS_END:
        case TokenType.SPACE:
            cond ~= token.value;
            break;
        case TokenType.LITERAL:
            cond ~= lit2squote(token.value);
            break;
        case TokenType.SYMBOL:
            if(token.value == "mod") {
                cond ~= "%";
                break;
            }
            break cond_loop;
        default:
            token.dump();
            throw new Exception("意図しないコードがあります");
        }
    }while(tokenStream.hasNext());
    if(token.type != TokenType.SYMBOL) throw new Exception("意図しない行終端です");
    if(token.value == "goto"){
        tokenStream.skipSpace();
        token = tokenStream.next();
        if(token.type != TokenType.GENERIC_LABEL) throw new Exception("ラベル以外は対応していません");
        return A.create(["href":"#" ~ token.value[1..$], "cond":std.string.strip(cond)]);
    }
    // TODO goto 以外に対応
    return new XNNop;
}

Command inputTokensToCommand(TokenStream tokenStream) {
    // input $VAR,STR,STR,NUM,NUM
    tokenStream.skipSpace();
    Token* token = tokenStream.next();
    string name;
    string title;
    string placeholder;
    string maxlength;
    
    if(token.type != TokenType.STRING_VARIABLE) throw new Exception("文字変数がありません");
    name = var2name(token.value);
    
    if(!tokenStream.skipToNextParam()) throw new Exception("次のパラメータがありません");
    token = tokenStream.next();
    if(token.type != TokenType.LITERAL) throw new Exception("文字列がありません");
    title = std.xml.encode(dequote(token.value));
    
    if(!tokenStream.skipToNextParam()) throw new Exception("次のパラメータがありません");
    token = tokenStream.next();
    if(token.type != TokenType.LITERAL) throw new Exception("文字列がありません");
    placeholder = std.xml.encode(dequote(token.value));
    
    if(!tokenStream.skipToNextParam()) throw new Exception("次のパラメータがありません");
    token = tokenStream.next();
    if(token.type != TokenType.NUMBER) throw new Exception("数値がありません");
    maxlength = token.value;
    
    if(!tokenStream.skipToNextParam()) throw new Exception("次のパラメータがありません");
    token = tokenStream.next();
    if(token.type != TokenType.NUMBER) throw new Exception("数値がありません");
    
    return Input.create(["name":name, "title":title, "placeholder":placeholder, "maxlength":maxlength]);
}

Command waveTokensToCommand(TokenStream tokenStream) {
    // wave STR
    // waveloop STR
    string loop = tokenStream.isPreviousValue("waveloop") ? XN_TRUE : XN_FALSE;
    tokenStream.skipSpace();
    Token* token = tokenStream.next();
    if(token.type != TokenType.LITERAL) throw new Exception("文字列以外は対応していません");
    string src = dequote(token.value);
    return Audio.create(["src":src, "loop":loop, "autoplay":XN_TRUE]);
}

Command dwaveTokensToCommand(TokenStream tokenStream) {
    // dwave NUM,STR
    // dwaveloop NUM,STR
    string loop = tokenStream.isPreviousValue("dwaveloop") ? XN_TRUE : XN_FALSE;
    tokenStream.skipSpace();
    Token* token = tokenStream.next();
    if(token.type != TokenType.NUMBER) throw new Exception("数値以外は対応していません");
    string id = token.value;

    if(!tokenStream.skipToNextParam()) throw new Exception("次のパラメータがありません");
    token = tokenStream.next();
    if(token.type != TokenType.LITERAL) throw new Exception("文字列以外は対応していません");
    string src = dequote(token.value);
    return Audio.create(["id":id, "src":src, "loop":loop, "autoplay":XN_TRUE]);
}
