module ns2xn;

private import std.xml;
private import std.stdio;
private import std.string;
private import std.conv;

private import nscr.token;
private import nscr.charset;
private import nscr.util;

private import option;
private import xnovel;

private Option options;

int convert(string nscript, ref Option opt){
    options = opt;
    Token[][] lines = parseToTokenLines(splitlines(fromNScrCharset(nscript)));
    auto builder = new XNovelBuilder();
    bool definition = true;
    bool contents = false;
    Command[] prevCommands;
    auto analyzeOnDefinition = (Token[] tokens){
        auto tokenStream = new TokenStream(tokens);
        if(!tokenStream.skipToAffectable()) return;
        auto token = tokenStream.next();
        // Definition not needed.
        if(token.type == TokenType.SYMBOL){
            switch(token.value){
            case "game":
                definition = false;
                break;
            case "stralias":
                straliasTokensToBuild(builder, tokenStream);
                break;
            case "numalias":
                numaliasTokensToBuild(builder, tokenStream);
                break;
            case "effect":
                effectTokensToBuild(builder, tokenStream);
                break;
            case "underline":
                underlineTokensToBuild(builder, tokenStream);
                break;
            default:
                // Unsupported command.
                break;
            }
        }
    };
    auto analyzeOnExecution = (Token[] tokens){
        auto tokenStream = new TokenStream(tokens);
        if(!tokenStream.skipToAffectable()) return;
        auto token = tokenStream.next();
        if(tokens[0].type == TokenType.SPECIAL_LABEL && tokens[0].value == "*start"){
            // Initialize for the execution.
            int base_pos = 1;
            if(options.underline >= Underline.BOTTOM){
                base_pos = 7;
            }
            else if(options.underline >= Underline.MIDDLE){
                base_pos = 4;
            }
            underlinedPosTable["l"] = to!string(base_pos++);
            underlinedPosTable["c"] = to!string(base_pos++);
            underlinedPosTable["r"] = to!string(base_pos++);
            contents = true;
        }
    };
    auto analyzeOnContents = (Token[] tokens){
        // set nav/h1
        auto commands = tokensToCommands(tokens);
        if(commands.length){
            if(typeid(commands[0]) == typeid(Nav) && isTextCommands(prevCommands)){
                auto nav = cast(Nav)commands[0];
                //string[string];
                auto text = toText(prevCommands);
                if(text.length) {
                    nav._title = H1.create(null, text);
                    builder.back(prevCommands.length);
                }
                else {
                    nav._title = H1.create(null, options.caption);
                }
            }
            builder.add(commands);
        }
        prevCommands = commands;
    };
    // analyze
    foreach(lineno, tokens; lines){
        try{
            if(tokens.length == 0) {
                if(contents) builder.add(new Blank);
                continue;
            }
            if(definition){
                analyzeOnDefinition(tokens);
            }
            else{
                if(!contents){
                    analyzeOnExecution(tokens);
                }
                else{
                    analyzeOnContents(tokens);
                }
            }
        }
        catch(Exception e){
            writeln("line " ~ to!string(lineno + 1) ~ ": " ~ toNScrCharset(e.msg));
            foreach(token; tokens){
                write(toNScrCharset(token.value));
            }
            return 10;
        }
        debug{
            foreach(token; tokens){
                write(toNScrCharset(token.value));
            }
            writeln("");
        }
    }
    XNovel xn = builder.create();
    string xnxml = xn.toXML();
    if(options.print){
        writeln(toNScrCharset(xnxml));
    }
    else{
        std.file.write(options.output, xnxml);
    }
    return 0;
}

string toText(Command[] commands){
    string text;
    if(commands is null) return null;
    foreach(command; commands){
        if(!isTextType(command)) break;
        text ~= join(command.pretty(0), "");
    }
    return text;
}

bool isTextCommands(Command[] commands){
    int index = 0;
    int last = commands.length-1; // The last is Br.
    while(index < last){
        if(!isTextType(commands[index++]))
            return false;
    }
    return true;
}

bool isTextType(Command command){
    auto type = typeid(command);
    return type == typeid(xnovel.Text) || type == typeid(Ruby);
}

Command[] tokensToCommands(Token[] tokens){
    Command[] commands;
    TokenStream tokenStream = new TokenStream(tokens);
    try {
        while(tokenStream.hasNext()){
            Token* token = tokenStream.next();
            switch(token.type){
            case TokenType.SYMBOL:
                switch(token.value){
                case "bg":
                    commands ~= bgTokensToCommand(tokenStream);
                    break;
                case "ld":
                    commands ~= ldTokensToCommand(tokenStream);
                    break;
                case "cl":
                    commands ~= clTokensToCommand(tokenStream);
                    break;
                case "bgm":
                    commands ~= bgmTokensToCommand(tokenStream);
                    break;
                case "select":
                    commands ~= selectTokensToCommand(tokenStream);
                    break;
                case "bgmstop":
                    commands ~= new XNStopbgm;
                    break;
                case "end":
                    commands ~= new XNClose;
                    break;
                case "movie":
                    commands ~= movieTokensToCommand(tokenStream);
                    break;
                case "wait":
                    commands ~= waitTokensToCommand(tokenStream);
                    break;
                case "mov":
                    commands ~= movTokensToCommand(tokenStream);
                    break;
                case "if":
                    commands ~= ifTokensToCommand(tokenStream);
                    break;
                case "input":
                    commands ~= inputTokensToCommand(tokenStream);
                    break;
                case "wave":
                case "waveloop":
                    commands ~= waveTokensToCommand(tokenStream);
                    break;
                case "dwave":
                case "dwaveloop":
                    commands ~= dwaveTokensToCommand(tokenStream);
                    break;
                case "stop":
                    commands ~= new XNStopAllse;
                    break;
                case "inc":
                case "dec":
                    commands ~= incdecTokensToCommand(tokenStream);
                    break;
                case "goto":
                    commands ~= gotoTokensToCommand(tokenStream);
                    break;
                case "tablegoto":
                    commands ~= tablegotoTokensToCommand(tokenStream);
                    break;
                case "caption":
                    commands ~= captionTokensToCommand(tokenStream);
                    break;
                default:
                    // ignore
                }
                break;
            
            case TokenType.STRING:
                commands ~= new xnovel.Text(token.value);
                tokenStream.skipSpace();
                if(!tokenStream.hasNext() || tokenStream.isNextType(TokenType.COMMENT)){
                    commands ~= new Br;
                }
                break;
            
            case TokenType.RUBY:
                string[] rbrt = std.string.split(token.value[1..$-1], "/");
                commands ~= Ruby.create(rbrt[0], rbrt[1]);
                if(!tokenStream.hasNext() || tokenStream.isNextType(TokenType.COMMENT)){
                    commands ~= new Br;
                }
                break;
            
            case TokenType.COMMENT:
                // ';' always exists here.
                if(!options.nonop) commands ~= new XNNop(token.value[1..$]);
                break;
            
            case TokenType.STRING_VARIABLE:
            case TokenType.NUMBER_VARIABLE:
                commands ~= new xnovel.Text("#{" ~ var2name(token.value) ~ "}");
                if(!tokenStream.hasNext() || tokenStream.isNextType(TokenType.COMMENT)){
                    commands ~= new Br;
                }
                break;
            
            case TokenType.LINETOP_LABEL:
                commands ~= Br.create(["id":token.value[1..$]]);
                break;
            default:
                debug {
                    token.dump();
                    writeln("");
                }
                break;
            }
        }
    }
    catch(Exception e){
        debug writeln(toNScrCharset(e.toString()));
        throw e;
    }
    return commands;
}

enum Message : string{
    INVALID_PARAM = "パラメータが不正です",
    STRING_NOT_FOUND = "文字列がありません",
    NUMBER_NOT_FOUND = "数値がありません",
    LABEL_NOT_FOUND = "ラベルがありません",
    REQUIRED_FIRST_PARAM = "パラメータが必要です",
    REQUIRED_NEXT_PARAM = "次のパラメータが必要です",
}

string[][string] effects;
string[string] underlinedPosTable;

unittest {
    auto b = new XNovelBuilder();
    Token[] tokens;
    
    tokens = parseToTokens(" HOGE,\"String\"");
    straliasTokensToBuild(b, new TokenStream(tokens));
    assert("hoge" in straliases, join(straliases.keys));
    assert(straliases["hoge"] == "\"String\"");
    
    tokens = parseToTokens(" FUGA,100");
    numaliasTokensToBuild(b, new TokenStream(tokens));
    assert("fuga" in numaliases, join(numaliases.keys));
    assert(numaliases["fuga"] == "100");
    
    numaliases["effect_no"] = "2";
    tokens = parseToTokens(" EFFECT_NO,10,500");
    effectTokensToBuild(b, new TokenStream(tokens));
    assert("2" in effects, join(effects.keys));
    assert(effects["2"].length == 2);
    
    tokens = parseToTokens(" 40");
    underlineTokensToBuild(b, new TokenStream(tokens));
    assert(options.underline == 40, to!string(options.underline));
}

void straliasTokensToBuild(ref XNovelBuilder builder, TokenStream tokenStream){
    string name;
    string tokenValue;
    if(!tokenStream.getParam(
        TokenType.SYMBOL, &name,
        TokenType.LITERAL, &tokenValue))
        throw new Exception(Message.INVALID_PARAM);
    name = tolower(name);
    string value = dequote(tokenValue);
    switch(name){
    case "xn_head_title":
        builder.setTitle(value);
        break;
    case "xn_body_title":
        builder.setBodyTitle(value);
        break;
    case "xn_direction":
    case "xn_text_rotation":
    case "xn_support":
    case "xn_curl":
        builder.setSettings(name, value);
        break;
    default:
        auto suffix = chompPrefix(name, "xn_body_");
        if(suffix != name){
            builder.setBodyInfo(suffix, value);
        }
        else{
            // Not aliases to hack.
            straliases[name] = tokenValue;
        }
    }
}

void numaliasTokensToBuild(ref XNovelBuilder builder, TokenStream tokenStream){
    string name, value;
    if(!tokenStream.getParam(
        TokenType.SYMBOL, &name,
        TokenType.NUMBER, &value))
        throw new Exception(Message.INVALID_PARAM);
    numaliases[tolower(name)] = value;
}

void effectTokensToBuild(ref XNovelBuilder builder, TokenStream tokenStream){
    string def, org, millis;
    if(!tokenStream.getParam(
        TokenType.NUMBER, &def,
        TokenType.NUMBER, &org,
        TokenType.NUMBER, &millis))
        throw new Exception(Message.INVALID_PARAM);
    effects[def] ~= [org, millis];
}

void underlineTokensToBuild(ref XNovelBuilder builder, TokenStream tokenStream){
    string underline;
    if(!tokenStream.getParam(
        TokenType.NUMBER, &underline))
        throw new Exception(Message.INVALID_PARAM);
    options.underline = to!int(underline);
}

/**
 * Detect whether type T is an array except string.
 */
template isArray(T){
    enum bool isArray = !is(T == string) && std.traits.isArray!(T);
}

class TokenStream {
    private Token[] _tokens;
    private int _index;
    private int _mark = -1;
    
    this(Token[] tokens){
        _tokens = tokens;
        _index = 0;
    }
    
    bool hasNext(){
        return _index < _tokens.length;
    }
    
    Token* next(){
        return &_tokens[_index++];
    }
    
    string nextValueAs(TokenType type){
        Token* token = &_tokens[_index++];
        return reduce(type, token);
    }
    
    void skipSpace(){
        while(_index < _tokens.length && _tokens[_index].type == TokenType.SPACE)
            _index++;
    }
    
    bool skipToAffectable(){
        skipSpace();
        if(!hasNext()) return false;
        return _tokens[_index].type != TokenType.COMMENT;
    }
    
    bool skipToFirstParam(){
        if(!skipToAffectable()) return false;
        return isAcceptableAsParam(_tokens[_index].type);
    }
    
    bool skipToNextParam(){
        skipSpace();
        if(!hasNext()) return false;
        if(_tokens[_index].type == TokenType.ASCII && _tokens[_index].value == ","){
            _index++;
            skipSpace();
            if(!hasNext()) return false;
            return isAcceptableAsParam(_tokens[_index].type);
        }
        return false;
    }
    
    bool isNextType(TokenType type){
        return _tokens[_index].type == type;
    }
    
    bool isPreviousValue(string value){
        return _tokens[_index - 1].value == value;
    }
    
    bool getParam(T...)(T params){
        static if(params.length){
            TokenType type = params[0];
            auto receiver = params[1];
            if(!skipToFirstParam()) return false;
            auto value = reduce(type, next());
            if(value is null) return false;
            handleParam(type, receiver, value);
            static if(isArray!(typeof(*receiver))) return getNextParam(params);
            else static if(params.length == 2) return true;
            else return getNextParam(params[2..$]);
        }
        else{
            if(!skipToFirstParam()) return true;
            do{
                next();
            }while(skipToNextParam());
            return false;
        }
    }
    
    private bool getNextParam(T...)(T params){
        TokenType type = params[0];
        auto receiver = params[1];
        static assert(!(params.length > 2 && isArray!(typeof(*receiver))), "Array Must Be Tailed");
        if(!skipToNextParam()) {
            // True if the receiver is an tailed array containing something.
            static if(isArray!(typeof(*receiver))) return receiver.length > 0;
            else return false;
        }
        auto value = reduce(type, next());
        if(value is null) return false;
        handleParam(type, receiver, value);
        // Continuing to get params or not.
        static if(isArray!(typeof(*receiver))) return getNextParam(params);
        else static if(params.length == 2) return true;
        else return getNextParam(params[2..$]);
    }
    
    private void handleParam(R)(TokenType tokenType, R receiver, string value){
        static if(is(typeof(*receiver) == string)){
            // string
            *receiver = to!(typeof(*receiver))(value);
        }
        else static if(std.traits.isArray!(typeof(*receiver))){
            // array
            *receiver ~= [to!(typeof(*receiver[0]))(value)];
        }
        else{
            static assert(false, typeof(receiver).stringof ~ "Not Supported Type");
        }
    }
    
    private string reduce(TokenType toType, Token* token){
        TokenType fromType = token.type;
        string value = token.value;
        if(toType == fromType) return value;
        debug writeln("reduce ", value, " from ", fromType, " to ", toType);
        switch(toType){
        case TokenType.NUMBER:
            if(fromType == TokenType.SYMBOL) return alias2num(value);
            return null;
        case TokenType.LITERAL:
            if(fromType == TokenType.SYMBOL) return alias2str(value);
            return null;
        case TokenType.GENERIC_LABEL:
            if(fromType == TokenType.SYMBOL) return dequote(alias2str(value));
            return null;
        default:
            return null;
        }
    }
    
    void mark(){
        _mark = _index;
    }
    
    void reset(){
        if(_mark == -1) throw new Exception("Not Marked Yet!");
        _index = _mark;
    }
}

unittest {
    writeln("Test Parser");
    string s = cast(string)std.file.read("main.html");
    check(s);
    auto xn = new XNovel(s);
    assert(xn._head._title == "famibee XNovel紹介");
    assert(xn._body._title == "XNovel紹介");
    Contents contents = xn._body._contents;
    Command[] commands = contents._commands;
    assert(commands.length == 2, to!string(commands.length));
    assert(commands[1].classinfo.name == "xnovel.XNStopbgm", commands[1].classinfo.name);
    P p = cast(P)(commands[0]);
    Command[] children = p._children;
    assert(children.length > 0);
    int order = 0;
    assert(typeid(children[order++]) == typeid(Img), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Fg), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Audio), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(xnovel.Text), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Br), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Nav), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Br), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Ruby), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(xnovel.Text), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Br), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(XNClose), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(XNWait), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(Video), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(XNLet), children[order].classinfo.name);
    assert(typeid(children[order++]) == typeid(XNTrace), children[order].classinfo.name);
    Img img = cast(Img)(children[0]);
    assert(img.id == "bg", img.id);
    assert(img.src == "bg/machi.jpg");
    Fg fg = cast(Fg)(children[1]);
    assert(fg.id == "0");
    assert(fg.src == "fg/usagi01.png");
    assert(fg.tate_pos == 2);
    assert(fg.yoko_pos == 8);
    Audio audio = cast(Audio)(children[2]);
    assert(audio.src == "bgm/music.mp3");
    assert(audio.loop);
    assert(!audio.autoplay);
    auto text = cast(xnovel.Text)(children[3]);
    assert(text.text == "うさぎちゃん「このアプリはどんなアプリなの？」", text.text);
}

unittest
{
    writeln("Test Builder");
    auto builder = new XNovelBuilder();
    auto xn = builder
                .setTitle("Builder Test")
                .setBodyTitle("Title")
                .setBodyInfo("author", "Author")
                .add(Img.create(["id":"bg", "src":"bg/hoge.jpg"]))
                .add(new xnovel.Text("ほげ"))
                .create();
    check(xn.toXML());
}

unittest {
    assert(millis2seconds("3000") == "3", millis2seconds("3000"));
    assert(millis2seconds("500") == "0.5", millis2seconds("500"));
    assert(millis2seconds("1") == "0.001", millis2seconds("1"));
    assert(var2name("%0") == "save:i0", var2name("%0"));
    assert(var2name("$0") == "save:s0", var2name("$0"));
    assert(var2name("%200") == "sys:i200", var2name("%200"));
    assert(var2name("$200") == "sys:s200", var2name("$200"));
    numaliases["hoge"] = "100";
    assert(var2name("%HOGE") == "save:i100", var2name("%HOGE"));
    assert(var2name("$HOGE") == "save:s100", var2name("$HOGE"));
    assert(lit2squoted("\"<ほげ>&\"") == "'&lt;ほげ&gt;&amp;'", lit2squoted("\"<ほげ>&\""));
    assert(lit2dequoted("\"<ほげ>&\"") == "&lt;ほげ&gt;&amp;", lit2dequoted("\"<ほげ>&\""));
    effects["4"] = ["10", "1000"];
    assert(timeOfUserEffect("4") == "1000");
}

unittest {
    underlinedPosTable = ["l":"7", "c":"8", "r":"9"];
    Token[] tokens;
    Command[] commands;
    
    Img img;
    tokens = parseToTokens("bg \"bg/hoge.jpg\",1");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Img));
    img = cast(Img)commands[0];
    assert(img.id == "bg");
    assert(img.src == "bg/hoge.jpg", img.src);

    Fg fg;
    tokens = parseToTokens("bg \"bg/hoge.jpg\",1,500");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Fg));
    fg = cast(Fg)commands[0];
    assert(fg.id == "bg");
    assert(fg.src == "bg/hoge.jpg", fg.src);
    assert(fg.time == "0.5", fg.time);
    
    effects["4"] = ["10", "1000"];
    tokens = parseToTokens("bg \"bg/hoge.jpg\",4");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Fg));
    fg = cast(Fg)commands[0];
    assert(fg.id == "bg");
    assert(fg.src == "bg/hoge.jpg", fg.src);
    assert(fg.time == "1", fg.time);

    tokens = parseToTokens("ld l,\"fg/hoge.png\",1");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Fg));
    fg = cast(Fg)commands[0];
    assert(fg.id == "0");
    assert(fg.src == "fg/hoge.png", fg.src);
    assert(fg.tate_pos == 1, to!string(fg.tate_pos));
    assert(fg.yoko_pos == 7, to!string(fg.yoko_pos));

    tokens = parseToTokens("ld l,\"fg/hoge.png\",10,500");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Fg));
    fg = cast(Fg)commands[0];
    assert(fg.id == "0");
    assert(fg.src == "fg/hoge.png", fg.src);
    assert(fg.tate_pos == 1, to!string(fg.tate_pos));
    assert(fg.yoko_pos == 7, to!string(fg.yoko_pos));
    assert(fg.time == "0.5", fg.time);

    tokens = parseToTokens("ld l,\"fg/hoge.png\",4");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Fg));
    fg = cast(Fg)commands[0];
    assert(fg.id == "0");
    assert(fg.src == "fg/hoge.png", fg.src);
    assert(fg.tate_pos == 1, to!string(fg.tate_pos));
    assert(fg.yoko_pos == 7, to!string(fg.yoko_pos));
    assert(fg.time == "1", fg.time);

    tokens = parseToTokens("cl l,1");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Fg));
    fg = cast(Fg)commands[0];
    assert(fg.id == "0");
    assert(fg.src is null, fg.src);

    tokens = parseToTokens("cl l,10,500");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Fg));
    fg = cast(Fg)commands[0];
    assert(fg.id == "0");
    assert(fg.src is null, fg.src);
    assert(fg.time == "0.5", fg.time);

    tokens = parseToTokens("cl l,4");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Fg));
    fg = cast(Fg)commands[0];
    assert(fg.id == "0");
    assert(fg.src is null, fg.src);
    assert(fg.time == "1", fg.time);

    tokens = parseToTokens("cl a,1");
    commands = tokensToCommands(tokens);
    assert(commands.length == 3, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Fg));
    fg = cast(Fg)commands[0];
    assert(fg.id == "0");
    assert(fg.src is null, fg.src);
    assert(typeid(commands[1]) == typeid(Fg));
    fg = cast(Fg)commands[1];
    assert(fg.id == "1");
    assert(fg.src is null, fg.src);
    assert(typeid(commands[2]) == typeid(Fg));
    fg = cast(Fg)commands[2];
    assert(fg.id == "2");
    assert(fg.src is null, fg.src);

    Audio audio;
    tokens = parseToTokens("bgm \"bgm/hoge.mp3\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Audio));
    audio = cast(Audio)commands[0];
    assert(audio.src == "bgm/hoge.mp3", audio.src);
    assert(audio.loop);
    assert(audio.autoplay);
    assert(audio.id == "bgm", audio.id);
    
    xnovel.Text text;
    tokens = parseToTokens("ほげほげ");
    commands = tokensToCommands(tokens);
    assert(commands.length == 2, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(xnovel.Text));
    text = cast(xnovel.Text)commands[0];
    assert(text.text == "ほげほげ", text.text);
    assert(typeid(commands[1]) == typeid(Br));
    
//    tokens = parseToTokens("選択してください");
    tokens = parseToTokens("select \"先頭へ\",*sel0,\"中間へ\",*sel1,\"後半へ\",*sel2,\"最後へ\",*sel3");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Nav));
    auto nav = cast(Nav)commands[0];
    // TODO Test nav > h1
    assert(nav._links.length == 4, to!string(nav._links.length));
    assert(nav._links[0].text == "先頭へ");
    assert(nav._links[0].href == "#sel0");
    
    tokens = parseToTokens("*sel0");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Br));
    auto br = cast(Br)commands[0];
    assert(br.id == "sel0", br.id);
    
    Ruby ruby;
    tokens = parseToTokens("(先頭/せんとう)です。");
    commands = tokensToCommands(tokens);
    assert(commands.length == 3, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Ruby));
    ruby = cast(Ruby)commands[0];
    assert(ruby._rb == "先頭", ruby._rb);
    assert(ruby._rt == "せんとう", ruby._rt);
    assert(ruby._rps[0] == "(", ruby._rps[0]);
    assert(ruby._rps[1] == ")", ruby._rps[1]);
    assert(typeid(commands[1]) == typeid(xnovel.Text));
    text = cast(xnovel.Text)commands[1];
    assert(text.text == "です。", text.text);
    assert(typeid(commands[2]) == typeid(Br));
    
    tokens = parseToTokens("bgmstop");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNStopbgm));
    
    tokens = parseToTokens("end");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNClose));

    tokens = parseToTokens("movie \"video/hoge.mp4\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Video));
    auto video = cast(Video)commands[0];
    assert(video.src == "video/hoge.mp4", audio.src);

    tokens = parseToTokens("wait 3000");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNWait));
    auto xn_wait = cast(XNWait)commands[0];
    assert(xn_wait.time == "3", xn_wait.time);
    
    // Test breaking if terminated by ruby.
    tokens = parseToTokens("これで(最後/さいご)");
    commands = tokensToCommands(tokens);
    assert(commands.length == 3, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(xnovel.Text));
    text = cast(xnovel.Text)commands[0];
    assert(text.text == "これで", text.text);
    assert(typeid(commands[1]) == typeid(Ruby));
    ruby = cast(Ruby)commands[1];
    assert(ruby._rb == "最後", ruby._rb);
    assert(ruby._rt == "さいご", ruby._rt);
    assert(ruby._rps[0] == "(", ruby._rps[0]);
    assert(ruby._rps[1] == ")", ruby._rps[1]);
    assert(typeid(commands[2]) == typeid(Br));
    
    XNLet xn_let;
    // $VAR, STR
    tokens = parseToTokens("mov $0,\"local\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNLet));
    xn_let = cast(XNLet)commands[0];
    assert(xn_let.name == "save:s0", xn_let.name);
    assert(xn_let.text == "local", xn_let.text);
    assert(!xn_let.calc);
    // %VAR, NUM
    tokens = parseToTokens("mov %0,100");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNLet));
    xn_let = cast(XNLet)commands[0];
    assert(xn_let.name == "save:i0", xn_let.name);
    assert(xn_let.text == "100", xn_let.text);
    assert(xn_let.calc);
    
    tokens = parseToTokens("if %0 > 99 goto *q_easy");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(A));
    auto a = cast(A)commands[0];
    assert(a.href == "#q_easy", a.href);
    assert(a.cond == "save:i0 ＞ 99", a.cond);
    
    tokens = parseToTokens("; 後ろ2つの数値は無視されます。");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNNop));
    auto xn_nop = cast(XNNop)commands[0];
    assert(xn_nop.text == " 後ろ2つの数値は無視されます。", xn_nop.text);
    
    tokens = parseToTokens("input $0,\"文字列を入力してください\",\"デフォルト\",24,0");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Input));
    auto input = cast(Input)commands[0];
    assert(input.title == "文字列を入力してください", input.title);
    assert(input.name == "save:s0", input.name);
    assert(input.placeholder == "デフォルト", input.placeholder);
    assert(input.maxlength == 24, to!string(input.maxlength));

    tokens = parseToTokens("変数の中身は$0");
    commands = tokensToCommands(tokens);
    assert(commands.length == 3, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(xnovel.Text));
    text = cast(xnovel.Text)commands[0];
    assert(text.text == "変数の中身は", text.text);
    assert(typeid(commands[1]) == typeid(xnovel.Text));
    text = cast(xnovel.Text)commands[1];
    assert(text.text == "#{save:s0}", text.text);
    assert(typeid(commands[2]) == typeid(Br));
    
    // wave
    tokens = parseToTokens("wave \"se/hoge.wav\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Audio));
    audio = cast(Audio)commands[0];
    assert(audio.src == "se/hoge.wav", audio.src);
    assert(!audio.loop);
    assert(audio.autoplay);
    assert(audio.id is null, audio.id);
    // waveloop
    tokens = parseToTokens("waveloop \"se/hoge.wav\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Audio));
    audio = cast(Audio)commands[0];
    assert(audio.src == "se/hoge.wav", audio.src);
    assert(audio.loop);
    assert(audio.autoplay);
    assert(audio.id is null, audio.id);
    // dwave
    tokens = parseToTokens("dwave 0,\"se/fuga.wav\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Audio));
    audio = cast(Audio)commands[0];
    assert(audio.src == "se/fuga.wav", audio.src);
    assert(!audio.loop);
    assert(audio.autoplay);
    assert(audio.id == "0", audio.id);
    // dwaveloop
    tokens = parseToTokens("dwaveloop 1,\"se/fuga.wav\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(Audio));
    audio = cast(Audio)commands[0];
    assert(audio.src == "se/fuga.wav", audio.src);
    assert(audio.loop);
    assert(audio.autoplay);
    assert(audio.id == "1", audio.id);
    
    tokens = parseToTokens("stop");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNStopAllse));
    
    tokens = parseToTokens("inc %0");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNLet));
    xn_let = cast(XNLet)commands[0];
    assert(xn_let.name == "save:i0", xn_let.name);
    assert(xn_let.text == "save:i0 + 1", xn_let.text);
    assert(xn_let.calc);
    
    tokens = parseToTokens("dec %0");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNLet));
    xn_let = cast(XNLet)commands[0];
    assert(xn_let.name == "save:i0", xn_let.name);
    assert(xn_let.text == "save:i0 - 1", xn_let.text);
    assert(xn_let.calc);
    
    tokens = parseToTokens("goto *q_easy");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(A));
    a = cast(A)commands[0];
    assert(a.href == "#q_easy", a.href);
    
    tokens = parseToTokens("tablegoto %0,*label1,*label2,*label3");
    commands = tokensToCommands(tokens);
    assert(commands.length == 3, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(A));
    a = cast(A)commands[0];
    assert(a.href == "#label1", a.href);
    assert(a.cond == "save:i0 == 0", a.cond);
    assert(typeid(commands[1]) == typeid(A));
    a = cast(A)commands[1];
    assert(a.href == "#label2", a.href);
    assert(a.cond == "save:i0 == 1", a.cond);
    assert(typeid(commands[2]) == typeid(A));
    a = cast(A)commands[2];
    assert(a.href == "#label3", a.href);
    assert(a.cond == "save:i0 == 2", a.cond);

    tokens = parseToTokens("caption \"caption\"");
    commands = tokensToCommands(tokens);
    assert(commands.length == 1, to!string(commands.length));
    assert(typeid(commands[0]) == typeid(XNNop));
    xn_nop = cast(XNNop)commands[0];
    assert(xn_nop.text == "caption", xn_nop.text);
}

string millis2seconds(string time){
    int milli = to!int(time);
    string sec = format("%1.3f", (milli / 1000.0f));
    auto m = std.regexp.search(sec, "0+$");
    if(!m) return sec;
    string zero = m.match(0);
    if(zero.length == 3) return sec[0..1];
    return sec[0..$-zero.length];
}

string var2name(string var){
    // エイリアスは受け付けない。
    string infix = var[0] == '$' ? "s" : "i";
    string suffix = var[1..$];
    if(suffix[0] == '%') throw new Exception("変数番号参照には対応していません");
    string num = alias2num(suffix);
    if(num.length) suffix = num;
    int varnum = to!int(suffix);
    return (varnum < 200 ? "save:" : "sys:") ~ infix ~ suffix;
}

string lit2squoted(string lit){
    string result = std.xml.encode(dequote(lit));
    return "'" ~ result ~ "'";
}

string lit2dequoted(string lit){
    return std.xml.encode(dequote(lit));
}

string label2href(string label){
    return "#" ~ label[1..$];
}

string timeOfUserEffect(string effno){
    if(effno in effects && effects[effno].length == 2){
        return effects[effno][1];
    }
    return null;
}

string yoko_posOnUnderline(string pos){
    if(pos in underlinedPosTable)
        return underlinedPosTable[pos];
    return null;
}

Command bgTokensToCommand(TokenStream tokenStream) {
    // bg STR,NUM
    string src, effno, time;
    tokenStream.mark();
    if(!tokenStream.getParam(
        TokenType.LITERAL, &src,
        TokenType.NUMBER, &effno,
        TokenType.NUMBER, &time))
    {
        tokenStream.reset();
        if(!tokenStream.getParam(
            TokenType.LITERAL, &src,
            TokenType.NUMBER, &effno))
            throw new Exception(Message.INVALID_PARAM);
    }
    src = dequote(src);
    if(time is null){
        time = timeOfUserEffect(effno);
    }
    if(time.length){
        // bg STR,NUM,NUM
        return Fg.create(["id":"bg", "src":src, "time":millis2seconds(time)]);
    }
    return Img.create(["id":"bg", "src":src]);
}

Command ldTokensToCommand(TokenStream tokenStream) {
    // ld NAME,STR,NUM
    string pos, src, effno, time;
    tokenStream.mark();
    if(!tokenStream.getParam(
        TokenType.SYMBOL, &pos,
        TokenType.LITERAL, &src,
        TokenType.NUMBER, &effno,
        TokenType.NUMBER, &time))
    {
        tokenStream.reset();
        if(!tokenStream.getParam(
            TokenType.SYMBOL, &pos,
            TokenType.LITERAL, &src,
            TokenType.NUMBER, &effno))
            throw new Exception(Message.INVALID_PARAM);
    }
    // 位置による振り分け
    string id;
    string tate_pos;
    string yoko_pos = yoko_posOnUnderline(pos);
    switch(pos){
    case "l":
        id = "0";
        tate_pos = "1";
        break;
    case "c":
        id = "1";
        tate_pos = "2";
        break;
    case "r":
        id = "2";
        tate_pos = "3";
        break;
    default:
        throw new Exception("不正な位置です:" ~ pos);
    }
    src = dequote(src);
    if(time is null){
        time = timeOfUserEffect(effno);
    }
    if(time.length){
        // ld NAME,STR,NUM,NUM
        return Fg.create(["id":id, "src":src, "tate_pos":tate_pos, "yoko_pos":yoko_pos, "time":millis2seconds(time)]);
    }
    return Fg.create(["id":id, "src":src, "tate_pos":tate_pos, "yoko_pos":yoko_pos]);
}

Command[] clTokensToCommand(TokenStream tokenStream) {
    // cl NAME,NUM
    string pos, effno, time;
    tokenStream.mark();
    if(!tokenStream.getParam(
        TokenType.SYMBOL, &pos,
        TokenType.NUMBER, &effno,
        TokenType.NUMBER, &time))
    {
        tokenStream.reset();
        if(!tokenStream.getParam(
            TokenType.SYMBOL, &pos,
            TokenType.NUMBER, &effno))
            throw new Exception(Message.INVALID_PARAM);
    }
    // 位置による振り分け
    string id;
    switch(pos){
    case "l":
        id = "0";
        break;
    case "c":
        id = "1";
        break;
    case "r":
        id = "2";
        break;
    case "a":
        id = null;
        break;
    default:
        throw new Exception("不正な位置です:" ~ pos);
    }
    if(time is null){
        time = timeOfUserEffect(effno);
    }
    if(id.length == 1){
        if(time.length){
            // cl NAME,NUM,NUM
            return [Fg.create(["id":id, "time":millis2seconds(time)])];
        }
        return [Fg.create(["id":id])];
    }
    else{
        if(time.length){
            // cl NAME,NUM,NUM
            string seconds = millis2seconds(time);
            return [
                Fg.create(["id":"0", "time":seconds]),
                Fg.create(["id":"1", "time":seconds]),
                Fg.create(["id":"2", "time":seconds])
            ];
        }
        return [Fg.create(["id":"0"]), Fg.create(["id":"1"]), Fg.create(["id":"2"])];
    }
}

Command bgmTokensToCommand(TokenStream tokenStream) {
    // bgm STR
    string src;
    if(!tokenStream.getParam(
        TokenType.LITERAL, &src))
        throw new Exception(Message.INVALID_PARAM);
    return Audio.create(["id":"bgm", "src":dequote(src), "loop":XN_TRUE, "autoplay":XN_TRUE]);
}

Command selectTokensToCommand(TokenStream tokenStream) {
    // select STR,LABEL,...
    auto nav = new Nav();
    if(!tokenStream.skipToFirstParam()) throw new Exception(Message.REQUIRED_FIRST_PARAM);
    do{
        string text = dequote(tokenStream.nextValueAs(TokenType.LITERAL));
        if(text is null) throw new Exception(Message.STRING_NOT_FOUND);
        
        if(!tokenStream.skipToNextParam()) throw new Exception(Message.REQUIRED_NEXT_PARAM);
        string label = tokenStream.nextValueAs(TokenType.GENERIC_LABEL);
        if(label is null) throw new Exception(Message.LABEL_NOT_FOUND);
        string href = "#" ~ label[1..$];
        nav._links ~= A.create(["href":href], text);
    }while(tokenStream.skipToNextParam());
    return nav;
}

Command movieTokensToCommand(TokenStream tokenStream) {
    // movie STR
    string src;
    if(!tokenStream.getParam(
        TokenType.LITERAL, &src))
        throw new Exception(Message.INVALID_PARAM);
    return Video.create(["src":dequote(src)]);
}

Command waitTokensToCommand(TokenStream tokenStream) {
    // wait NUM
    string time;
    if(!tokenStream.getParam(
        TokenType.NUMBER, &time))
        throw new Exception(Message.INVALID_PARAM);
    return XNWait.create(["time":millis2seconds(time)]);
}

Command movTokensToCommand(TokenStream tokenStream) {
    // mov VAR,STR|NUM
    tokenStream.skipSpace();
    Token* token = tokenStream.next();
    string name;
    string value;
    string calc;
    if(token.type == TokenType.NUMBER_VARIABLE) {
        name = var2name(token.value);
        calc = XN_TRUE;
        if(!tokenStream.skipToNextParam()) throw new Exception("意図しないコードがあります");
        // 計算式と見て、modの変換以外はすべて受け付ける。
        calc_loop:
        do{
            token = tokenStream.next();
            switch(token.type){
            case TokenType.NUMBER:
            case TokenType.PARENTHESIS_BEGIN:
            case TokenType.PARENTHESIS_END:
            case TokenType.SPACE:
                value ~= token.value;
                break;
            case TokenType.OPERATOR:
                if(!inPattern(token.value[0], "-+*/")) throw new Exception("意図しないコードがあります");
                value ~= token.value;
                break;
            case TokenType.SYMBOL:
                if(token.value == "mod"){
                    value ~= "%";
                }
                else{
                    string num = alias2num(token.value);
                    if(num.length) value ~= num;
                    else throw new Exception(Message.INVALID_PARAM);
                }
                break;
            case TokenType.COMMENT:
                break calc_loop;
            default:
                token.dump();
                throw new Exception("意図しないコードがあります");
            }
        }while(tokenStream.hasNext());
    }
    else if(token.type == TokenType.STRING_VARIABLE) {
        name = var2name(token.value);
        calc = XN_FALSE;
        if(!tokenStream.skipToNextParam()) throw new Exception("意図しないコードがあります");
        // 文字列結合と見て、modの変換以外はすべて受け付ける。
        cat_loop:
        do{
            token = tokenStream.next();
            switch(token.type){
            case TokenType.LITERAL:
                value ~= lit2dequoted(token.value);
                break;
            case TokenType.SPACE:
                value ~= token.value;
                break;
            case TokenType.SYMBOL:
                {
                    string str = alias2str(token.value);
                    if(str.length) value ~= lit2dequoted(str);
                    else throw new Exception(Message.INVALID_PARAM);
                }
                break;
            case TokenType.OPERATOR:
                // XNovel側で結合されるかはなぞ。
                if(token.value != "+") throw new Exception("意図しないコードがあります");
                calc = XN_TRUE;
                value ~= token.value;
                break;
            case TokenType.COMMENT:
                break cat_loop;
            default:
                token.dump();
                throw new Exception("意図しないコードがあります");
            }
        }while(tokenStream.hasNext());
    }
    else {
        throw new Exception("意図しないコードがあります");
    }
    return XNLet.create(["name":name, "text":value, "calc":calc]);
}

Command ifTokensToCommand(TokenStream tokenStream) {
    // if CONDITION COMMAND
    // COMMAND only "goto LABEL"
    tokenStream.skipSpace();
    Token* token;
    string cond;
    // COMMANDまでは条件句
    cond_loop:
    do{
        token = tokenStream.next();
        switch(token.type){
        case TokenType.NUMBER_VARIABLE:
        case TokenType.STRING_VARIABLE:
            cond ~= var2name(token.value);
            break;
        case TokenType.OPERATOR:
        case TokenType.ASCII:
            // TODO パターン検証
            // 実体参照だと戻されるので、とりあえず全角で対処
            cond ~= std.xml.encode(std.array.replace(std.array.replace(token.value, "<", "＜"), ">", "＞"));
            break;
        case TokenType.NUMBER:
        case TokenType.PARENTHESIS_BEGIN:
        case TokenType.PARENTHESIS_END:
        case TokenType.SPACE:
            cond ~= token.value;
            break;
        case TokenType.LITERAL:
            cond ~= lit2squoted(token.value);
            break;
        case TokenType.SYMBOL:
            if(token.value == "mod") {
                cond ~= "%";
                break;
            }
            else {
                string str = alias2str(token.value);
                string num = alias2num(token.value);
                if(str.length) cond ~= lit2squoted(str);
                else if(num.length) cond ~= num;
            }
            break cond_loop;
        default:
            token.dump();
            throw new Exception("意図しないコードがあります");
        }
    }while(tokenStream.hasNext());
    if(token.type != TokenType.SYMBOL) throw new Exception("意図しない行終端です");
    if(token.value == "goto"){
        return gotoTokensToCommand(tokenStream, strip(cond));
    }
    // TODO goto 以外に対応
    return new XNNop;
}

Command inputTokensToCommand(TokenStream tokenStream) {
    // input $VAR,STR,STR,NUM,NUM
    string name;
    string title;
    string placeholder;
    string maxlength;
    string zenkaku;
    if(!tokenStream.getParam(
        TokenType.STRING_VARIABLE, &name,
        TokenType.LITERAL, &title,
        TokenType.LITERAL, &placeholder,
        TokenType.NUMBER, &maxlength,
        TokenType.NUMBER, &zenkaku))
        throw new Exception(Message.INVALID_PARAM);
    
    name = var2name(name);
    title = std.xml.encode(dequote(title));
    placeholder = std.xml.encode(dequote(placeholder));
    
    return Input.create(["name":name, "title":title, "placeholder":placeholder, "maxlength":maxlength]);
}

Command waveTokensToCommand(TokenStream tokenStream) {
    // wave STR
    // waveloop STR
    string loop = tokenStream.isPreviousValue("waveloop") ? XN_TRUE : XN_FALSE;
    string src;
    if(!tokenStream.getParam(
        TokenType.LITERAL, &src))
        throw new Exception(Message.INVALID_PARAM);
    return Audio.create(["src":dequote(src), "loop":loop, "autoplay":XN_TRUE]);
}

Command dwaveTokensToCommand(TokenStream tokenStream) {
    // dwave NUM,STR
    // dwaveloop NUM,STR
    string loop = tokenStream.isPreviousValue("dwaveloop") ? XN_TRUE : XN_FALSE;
    string id, src;
    if(!tokenStream.getParam(
        TokenType.NUMBER, &id,
        TokenType.LITERAL, &src))
        throw new Exception(Message.INVALID_PARAM);
    return Audio.create(["id":id, "src":dequote(src), "loop":loop, "autoplay":XN_TRUE]);
}

Command incdecTokensToCommand(TokenStream tokenStream) {
    // inc %VAR
    // dec %VAR
    string operator = tokenStream.isPreviousValue("inc") ? "+" : "-";
    string var;
    if(!tokenStream.getParam(TokenType.NUMBER_VARIABLE, &var))
        throw new Exception(Message.INVALID_PARAM);
    string name = var2name(var);
    return XNLet.create(["name":name, "text":name ~ " " ~ operator ~ " 1"]);
}

Command gotoTokensToCommand(TokenStream tokenStream, string cond = null) {
    // goto LABEL
    string label;
    if(!tokenStream.getParam(TokenType.GENERIC_LABEL, &label))
        throw new Exception(Message.INVALID_PARAM);
    if(cond is null){
        return A.create(["href":label2href(label)]);
    }
    return A.create(["href":label2href(label), "cond":cond]);
}

Command[] tablegotoTokensToCommand(TokenStream tokenStream) {
    // tablegoto %VAR,LABEL,...
    string name;
    string[] labels;
    if(!tokenStream.getParam(
        TokenType.NUMBER_VARIABLE, &name,
        TokenType.GENERIC_LABEL, &labels))
        throw new Exception(Message.INVALID_PARAM);
    name = var2name(name);
    Command[] commands;
    foreach(index, label; labels){
        commands ~= A.create(["href":label2href(label), "cond":name ~ " == " ~ to!string(index)]);
    }
    return commands;
}

Command captionTokensToCommand(TokenStream tokenStream) {
    // caption STR
    string caption;
    if(!tokenStream.getParam(
        TokenType.LITERAL, &caption))
        throw new Exception(Message.INVALID_PARAM);
    if(caption.length){
        caption = dequote(caption);
        options.caption = caption;
    }
    if(options.nonop)
        return new Blank;
    return new XNNop(caption);
}
